#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# For live logging, use the following command:
#   pytest -o log_cli=true --log-cli-level=DEBUG test_optimizer_huggingface_bert.py

import shutil
import unittest
from pathlib import Path

import torch
from parity_utilities import find_transformers_source
from transformers.utils import default_cache_path

if find_transformers_source():
    from benchmark_helper import ConfigModifier, OptimizerInfo, Precision
    from compare_bert_results import run_test as bert_parity_test
    from onnx_exporter import export_onnx_model_from_pt
else:
    from onnxruntime.transformers.benchmark_helper import ConfigModifier, OptimizerInfo, Precision
    from onnxruntime.transformers.compare_bert_results import run_test as bert_parity_test
    from onnxruntime.transformers.onnx_exporter import export_onnx_model_from_pt


class TestHuggingfaceBertModelOptimization(unittest.TestCase):
    def run_optimizer_on_model(
        self,
        model_name,
        expected_fusion_result_list,
        inputs_count=1,
        validate_model=True,
        opset_version=16,
        use_external_data_format=False,
        model_type="bert",
    ):
        onnx_dir = Path(".") / "onnx_models" / model_name
        shutil.rmtree(onnx_dir, ignore_errors=True)

        Path(onnx_dir).mkdir(parents=True, exist_ok=True)

        model_fusion_statistics = {}

        input_names = ["input_ids", "attention_mask", "token_type_ids"]

        config_modifier = ConfigModifier(None)
        fusion_options = None
        model_class = "AutoModel"
        with torch.no_grad():
            optimized_model_path, is_valid_onnx_model, _, _ = export_onnx_model_from_pt(
                model_name=model_name,
                opset_version=opset_version,
                use_external_data_format=use_external_data_format,
                model_type=model_type,
                model_class=model_class,
                config_modifier=config_modifier,
                cache_dir=default_cache_path,
                onnx_dir=str(onnx_dir),
                input_names=input_names[:inputs_count],
                use_gpu=False,
                precision=Precision.FLOAT32,
                optimizer_info=OptimizerInfo.BYSCRIPT,
                validate_onnx=True,
                use_raw_attention_mask=True,
                overwrite=True,
                model_fusion_statistics=model_fusion_statistics,
                fusion_options=fusion_options,
            )

        if validate_model:
            self.assertEqual(is_valid_onnx_model, True)

        expected_node_count = {
            "EmbedLayerNormalization": expected_fusion_result_list[0],
            "Attention": expected_fusion_result_list[1],
            "Gelu": expected_fusion_result_list[2],
            "FastGelu": expected_fusion_result_list[3],
            "BiasGelu": expected_fusion_result_list[4],
            "LayerNormalization": expected_fusion_result_list[5],
            "SkipLayerNormalization": expected_fusion_result_list[6],
        }

        node_count = None
        for value in model_fusion_statistics.values():
            node_count = value
        self.assertIsNotNone(node_count)

        actual_node_count = {}
        for op_type in expected_node_count:
            actual_node_count[op_type] = node_count.get(op_type, 0)

        expected = ", ".join(f"{key}: {value}" for key, value in sorted(expected_node_count.items()))
        actual = ", ".join(f"{key}: {value}" for key, value in sorted(actual_node_count.items()))
        self.assertEqual(expected, actual)

        suffix = "_fp32_cpu.onnx"
        assert optimized_model_path.endswith(suffix)
        baseline_model_path = optimized_model_path[: -len(suffix)] + ".onnx"
        for batch_size in [1, 2]:
            for sequence_length in [1, 8]:
                max_abs_diff, case_passed = bert_parity_test(
                    baseline_model_path,
                    optimized_model_path,
                    output_dir=None,
                    batch_size=batch_size,
                    sequence_length=sequence_length,
                    use_gpu=False,
                    test_cases=1,
                    seed=123,
                    verbose=False,
                    rtol=1e-4,
                    atol=1e-4,
                    input_ids_name=input_names[0],
                    segment_ids_name=input_names[2] if inputs_count > 2 else None,
                    input_mask_name=input_names[1] if inputs_count > 1 else None,
                    mask_type=2,
                    dictionary_size=1024,
                )
                self.assertTrue(
                    case_passed, f"bert parity test failed: {batch_size=} {sequence_length=} {max_abs_diff=}"
                )

    def test_bert(self):
        model_name = "hf-internal-testing/tiny-random-bert"
        self.run_optimizer_on_model(model_name, [1, 5, 0, 0, 5, 0, 10], inputs_count=1)
        self.run_optimizer_on_model(model_name, [1, 5, 0, 0, 5, 0, 10], inputs_count=2)
        self.run_optimizer_on_model(model_name, [1, 5, 0, 0, 5, 0, 10], inputs_count=3)

    def test_roberta(self):
        model_name = "hf-internal-testing/tiny-random-roberta"
        # TODO: EmbedLayerNormalization fusion.
        self.run_optimizer_on_model(model_name, [0, 5, 0, 0, 5, 1, 10], inputs_count=1)
        self.run_optimizer_on_model(model_name, [0, 5, 0, 0, 5, 1, 10], inputs_count=2)

    def test_distillbert(self):
        model_name = "hf-internal-testing/tiny-random-distilbert"
        self.run_optimizer_on_model(model_name, [1, 5, 0, 0, 5, 0, 10], inputs_count=1)
        self.run_optimizer_on_model(model_name, [1, 5, 0, 0, 5, 0, 10], inputs_count=2)

    def test_xlm_roberta(self):
        model_name = "hf-internal-testing/tiny-xlm-roberta"
        # TODO: EmbedLayerNormalization fusion.
        self.run_optimizer_on_model(model_name, [0, 2, 0, 0, 2, 1, 4], inputs_count=1)
        self.run_optimizer_on_model(model_name, [0, 2, 0, 0, 2, 1, 4], inputs_count=2)


if __name__ == "__main__":
    unittest.main()



import onnx
from onnx import TensorProto, helper

graph = helper.make_graph(
    [  # nodes
        helper.make_node("Add", ["A", "B"], ["C"], "Add"),
    ],
    "SingleAdd",  # name
    [  # inputs
        helper.make_tensor_value_info("A", TensorProto.FLOAT, [1]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, [1]),
    ],
    [  # outputs
        helper.make_tensor_value_info("C", TensorProto.FLOAT, [1]),
    ],
)

model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 12)])
onnx.save(model, r"single_add.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------


import os
from dataclasses import dataclass
from itertools import product

import kernel_explorer as ke
import numpy as np
import pytest
from utils import dtype_to_suffix, matmul, softmax

max_batch_size = int(os.environ.get("KERNEL_EXPLORER_BATCHED_GEMM_MAX_BATCH_SIZE", 64))


def multinormal_distribution(num_distribution, num_element_per_dist):
    arrays = []
    for _ in range(num_distribution):
        mean = np.random.rand() - 0.5
        std = np.random.rand() + 0.5
        arrays.append(np.random.normal(mean, std, (num_element_per_dist,)))
    return np.array(arrays)


def get_ck_binding_name(dtype, biased: bool, masked: bool):
    dtype_suffix = "_" + dtype_to_suffix(dtype)
    ck_suffix = ""
    if biased:
        ck_suffix += "Biased"
    if masked:
        ck_suffix += "Masked"
    ck_suffix += dtype_suffix
    return "GemmSoftmaxGemmPermuteCK" + ck_suffix


dtypes = ["float16"]
batches = [1, max_batch_size]
seqlens = [128, 512]
total_seqlens = [128, 512]
num_heads = [8, 12]
head_sizes = [64]
biaseds = [False, True]
causals = [False]
mask_dims = [0, 2, 3, 4]


def get_biased_id(biased):
    return "biased" if biased else "nobias"


def get_mask_dim_id(dim):
    if dim == 0:
        return "nomask"
    return f"mask_{dim}d"


def maybe_pack_q_k_v_bnsh_for_device_on_host(q, k, v, dtype, qkv_format):
    q = q.astype(dtype)
    k = k.astype(dtype)
    v = v.astype(dtype)
    if qkv_format == ke.qkv_format.Q_K_V_BNSH:
        return q, k, v

    # BNSH to BSNH
    q = np.swapaxes(q, 2, 1)
    k = np.swapaxes(k, 2, 1)
    v = np.swapaxes(v, 2, 1)

    if qkv_format == ke.qkv_format.Q_K_V_BSNH:
        return np.ascontiguousarray(q), np.ascontiguousarray(k), np.ascontiguousarray(v)

    if qkv_format == ke.qkv_format.QKV_BSN3H:
        return np.ascontiguousarray(np.stack([q, k, v], axis=-2)), None, None

    if qkv_format == ke.qkv_format.Q_KV_BSNH_BSN2H:
        return np.ascontiguousarray(q), np.ascontiguousarray(np.stack([k, v], axis=-2)), None

    raise NotImplementedError


def _make_causal_mask(
    seqence_length,
    total_sequence_length,
    dtype: np.dtype,
):
    """
    Make causal mask used for Attention with attribute unidirectional == 1.
    The mask is a upper triangular matrix with shape [sequence_length, total_sequence_length].
    Putting a 1 indicates that the token at this position should be masked.
    For Example:
    sequence_length = 5, total_sequence_length = 5,
    mask: [[0. 1. 1. 1. 1.]
           [0. 0. 1. 1. 1.]
           [0. 0. 0. 1. 1.]
           [0. 0. 0. 0. 1.]
           [0. 0. 0. 0. 0.]]
    seqence_length = 5, total_seqence_length = 3,
    mask: [[1. 1. 1.]
           [1. 1. 1.]
           [0. 1. 1.]
           [0. 0. 1.]
           [0. 0. 0.]]
    seqence_length = 5, total_seqence_length = 7,
    mask: [[0. 0. 0. 1. 1. 1. 1.]
           [0. 0. 0. 0. 1. 1. 1.]
           [0. 0. 0. 0. 0. 1. 1.]
           [0. 0. 0. 0. 0. 0. 1.]
           [0. 0. 0. 0. 0. 0. 0.]]
    """
    mask = np.full((seqence_length, seqence_length), 1)
    mask_cond = np.arange(mask.shape[-1])
    mask = np.where(mask_cond < (mask_cond + 1).reshape(mask.shape[-1], 1), 0, mask)

    mask = mask.astype(dtype)

    if total_sequence_length - seqence_length > 0:
        mask = np.concatenate(
            [np.zeros((seqence_length, total_sequence_length - seqence_length), dtype=dtype), mask], axis=-1
        )

    if total_sequence_length - seqence_length < 0:
        mask = mask[:, -total_sequence_length:]

    correct_mask = np.full((seqence_length, total_sequence_length), 1)
    for i in range(seqence_length):
        correct_mask[i][:] = sum(mask[i]) != total_sequence_length
    return mask, correct_mask


def _test_gemm_softmax_gemm_permute(
    f, dtype, batch, seqlen, total_seqlen, num_heads, head_size, biased, mask_dim, scale, causal, qkv_format
):
    v_head_size = head_size
    q_shape = [batch, num_heads, seqlen, head_size]
    k_shape = [batch, num_heads, total_seqlen, head_size]
    v_shape = [batch, num_heads, total_seqlen, v_head_size]
    out_shape = [batch, seqlen, num_heads, head_size]

    attn_bias = None
    bias_shape = [batch, num_heads, seqlen, total_seqlen] if biased else None

    attn_mask = None
    mask_shape = None
    mask_shape_broadcasted = None
    max_seqlen = None
    if mask_dim != 0:
        if mask_dim == 2:
            mask_shape = [batch, total_seqlen]
            mask_shape_broadcasted = [batch, 1, 1, total_seqlen]
        elif mask_dim == 3:
            mask_shape = [batch, seqlen, total_seqlen]
            mask_shape_broadcasted = [batch, 1, seqlen, total_seqlen]
        elif mask_dim == 4:
            max_seqlen = ((seqlen - 1) // 1024 + 1) * 1024  # round up to multiple of 1024
            mask_shape = [batch, 1, max_seqlen, max_seqlen]
        else:
            raise ValueError

    np.random.seed(42)
    q = multinormal_distribution(np.prod(q_shape[:-1]), q_shape[-1]).reshape(q_shape).astype(np.float64)
    k = multinormal_distribution(np.prod(k_shape[:-1]), k_shape[-1]).reshape(k_shape).astype(np.float64)
    v = multinormal_distribution(np.prod(v_shape[:-1]), v_shape[-1]).reshape(v_shape).astype(np.float64)
    if bias_shape is not None:
        attn_bias = np.random.uniform(-0.5, 0.5, size=bias_shape)
    if mask_shape is not None:
        attn_mask = (np.random.randint(0, 100, size=mask_shape) < 95).astype(np.int32)

    pre_softmax_attn_scores = matmul(q, np.swapaxes(k, 2, 3))
    pre_softmax_attn_scores = pre_softmax_attn_scores * scale
    if attn_bias is not None:
        pre_softmax_attn_scores = pre_softmax_attn_scores + attn_bias

    correct_causal_mask = np.full((seqlen, total_seqlen), 1)
    if attn_mask is not None:
        filter_value = -10000.0
        if mask_dim == 4:
            # equivalent to past_sequence_length = max_sequence_length - seqlen
            converted_mask = (1 - attn_mask[:, :, -seqlen:, :total_seqlen]) * filter_value
        else:
            converted_mask = (1 - attn_mask.reshape(mask_shape_broadcasted)) * filter_value
        pre_softmax_attn_scores = pre_softmax_attn_scores + converted_mask
    if causal:
        filter_value = np.finfo(dtype).min
        causal_mask, correct_causal_mask = _make_causal_mask(seqlen, total_seqlen, pre_softmax_attn_scores.dtype)
        causal_mask = np.broadcast_to(causal_mask, pre_softmax_attn_scores.shape) * filter_value
        pre_softmax_attn_scores = pre_softmax_attn_scores + causal_mask
    attn_scores = softmax(pre_softmax_attn_scores, axis=-1)

    # apply mask to attn_scores to correct softmax result, in c++ implementation, if all values in a row are masked,
    # the softmax result in this row will be filled with 0.
    correct_causal_mask = np.broadcast_to(correct_causal_mask, pre_softmax_attn_scores.shape)
    attn_scores = attn_scores * correct_causal_mask

    attn = matmul(attn_scores, v)
    ref = np.swapaxes(attn, 2, 1)  # permute 0213

    out = np.empty(out_shape, dtype=dtype)
    host_q, host_k, host_v = maybe_pack_q_k_v_bnsh_for_device_on_host(q, k, v, dtype, qkv_format)
    host_attn_bias = attn_bias.astype(dtype) if attn_bias is not None else None
    dev_q = ke.DeviceArray(host_q)
    dev_k = ke.DeviceArray(host_k) if host_k is not None else None
    dev_v = ke.DeviceArray(host_v) if host_v is not None else None
    dev_out = ke.DeviceArray(out)
    dev_attn_bias = ke.DeviceArray(host_attn_bias) if host_attn_bias is not None else None
    dev_attn_mask = ke.DeviceArray(attn_mask) if attn_mask is not None else None

    my_gemm_softmax_gemm_permute = f(
        batch,
        seqlen,
        total_seqlen,
        max_seqlen,
        num_heads,
        head_size,
        mask_dim,
        scale,
        causal,
        qkv_format,
        dev_q,
        dev_k,
        dev_v,
        dev_attn_bias,
        dev_attn_mask,
        dev_out,
    )

    print()  # write an empty line in case pytest ... -s -v
    failures = {}
    for impl in my_gemm_softmax_gemm_permute.ListOps():
        if not my_gemm_softmax_gemm_permute.SelectOp(impl):
            print("Unsupport", impl)
            continue
        print("  Support", impl)

        my_gemm_softmax_gemm_permute.Run()
        dev_out.UpdateHostNumpyArray()

        try:
            is_strict = int(os.environ.get("KERNEL_EXPLORER_STRICT_TEST", "0"))
            if is_strict:
                # NOTE: this will always fail, just for manual checking with:
                #  KERNEL_EXPLORER_STRICT_TEST=1 pytest ... -s -v
                np.testing.assert_allclose(out, ref)
            else:
                is_zero_tol, atol, rtol = 1e-3, 2e-2, 1e-2
                not_close_to_zeros = np.abs(ref) > is_zero_tol
                np.testing.assert_allclose(out[not_close_to_zeros], ref[not_close_to_zeros], atol=atol, rtol=rtol)
        except Exception as err:
            header = "*" * 30 + impl + "*" * 30
            print(header)
            print(err)
            print("*" * len(header))
            failures[impl] = str(err)

    if failures:
        raise Exception(failures)


@pytest.mark.parametrize("mask_dim", mask_dims, ids=get_mask_dim_id)
@pytest.mark.parametrize("biased", biaseds, ids=get_biased_id)
@pytest.mark.parametrize("head_size", head_sizes)
@pytest.mark.parametrize("nhead", num_heads)
@pytest.mark.parametrize("total_seqlen", total_seqlens)
@pytest.mark.parametrize("seqlen", seqlens)
@pytest.mark.parametrize("batch", [16])
@pytest.mark.parametrize("causal", [False, True])
@pytest.mark.parametrize("dtype", ["float16", "float32"])
def test_gemm_softmax_gemm_permute_generic(
    dtype, batch, seqlen, total_seqlen, nhead, head_size, biased, causal, mask_dim
):
    f = getattr(ke, "GemmSoftmaxGemmPermuteGeneric_" + dtype_to_suffix(dtype))
    scale = 1.0 / np.sqrt(head_size)
    _test_gemm_softmax_gemm_permute(
        f,
        dtype,
        batch,
        seqlen,
        total_seqlen,
        nhead,
        head_size,
        biased,
        mask_dim,
        scale,
        causal,
        ke.qkv_format.Q_K_V_BNSH,
    )


@pytest.mark.parametrize("mask_dim", [2], ids=get_mask_dim_id)
@pytest.mark.parametrize("biased", [False], ids=get_biased_id)
@pytest.mark.parametrize("head_size", [64])
@pytest.mark.parametrize("nhead", [8])
@pytest.mark.parametrize("total_seqlen", [128])
@pytest.mark.parametrize("seqlen", [64])
@pytest.mark.parametrize("batch", [16])
@pytest.mark.parametrize("causal", [True, False])
@pytest.mark.parametrize("dtype", ["float16", "float32"])
def test_gemm_softmax_gemm_permute_generic_nested_tunable(
    dtype, batch, seqlen, total_seqlen, nhead, head_size, biased, causal, mask_dim
):
    f = getattr(ke, "GemmSoftmaxGemmPermuteGenericNestedTunable_" + dtype_to_suffix(dtype))
    scale = 1.0 / np.sqrt(head_size)
    _test_gemm_softmax_gemm_permute(
        f,
        dtype,
        batch,
        seqlen,
        total_seqlen,
        nhead,
        head_size,
        biased,
        mask_dim,
        scale,
        causal,
        ke.qkv_format.Q_K_V_BNSH,
    )


@pytest.mark.skipif(not ke.is_composable_kernel_available(), reason="ck is not enabled")
@pytest.mark.parametrize("mask_dim", mask_dims, ids=get_mask_dim_id)
@pytest.mark.parametrize("biased", biaseds, ids=get_biased_id)
@pytest.mark.parametrize("head_size", head_sizes)
@pytest.mark.parametrize("nhead", num_heads)
@pytest.mark.parametrize("total_seqlen", total_seqlens)
@pytest.mark.parametrize("seqlen", seqlens)
@pytest.mark.parametrize("batch", batches)
@pytest.mark.parametrize("causal", [False, True])
@pytest.mark.parametrize("dtype", dtypes)
def test_gemm_softmax_gemm_permute_ck(dtype, batch, seqlen, total_seqlen, nhead, head_size, biased, causal, mask_dim):
    f = getattr(ke, get_ck_binding_name(dtype, biased, mask_dim != 0))
    scale = 1.0 / np.sqrt(head_size)
    _test_gemm_softmax_gemm_permute(
        f,
        dtype,
        batch,
        seqlen,
        total_seqlen,
        nhead,
        head_size,
        biased,
        mask_dim,
        scale,
        causal,
        ke.qkv_format.Q_K_V_BNSH,
    )


@pytest.mark.parametrize("mask_dim", [2], ids=get_mask_dim_id)
@pytest.mark.parametrize("biased", [False], ids=get_biased_id)
@pytest.mark.parametrize("head_size", [64])
@pytest.mark.parametrize("nhead", [8])
@pytest.mark.parametrize("total_seqlen", [128])
@pytest.mark.parametrize("seqlen", [64])
@pytest.mark.parametrize("batch", [16])
@pytest.mark.parametrize("causal", [False, True])
@pytest.mark.parametrize("dtype", ["float16"])
def test_gemm_softmax_gemm_permute_tunable(
    dtype, batch, seqlen, total_seqlen, nhead, head_size, biased, causal, mask_dim
):
    f = getattr(ke, "GemmSoftmaxGemmPermuteTunable_" + dtype_to_suffix(dtype))
    scale = 1.0 / np.sqrt(head_size)
    _test_gemm_softmax_gemm_permute(
        f,
        dtype,
        batch,
        seqlen,
        total_seqlen,
        nhead,
        head_size,
        biased,
        mask_dim,
        scale,
        causal,
        ke.qkv_format.Q_K_V_BNSH,
    )


stabel_diffusion_configs = [
    [2, 64, 64, 8, 160, "QKV_BSN3H"],
    [2, 256, 256, 8, 160, "QKV_BSN3H"],
    [2, 1024, 1024, 8, 80, "QKV_BSN3H"],
    [2, 4096, 4096, 8, 40, "QKV_BSN3H"],
    [2, 64, 77, 8, 160, "Q_KV_BSNH_BSN2H"],
    [2, 256, 77, 8, 160, "Q_KV_BSNH_BSN2H"],
    [2, 1024, 77, 8, 80, "Q_KV_BSNH_BSN2H"],
    [2, 4096, 77, 8, 40, "Q_KV_BSNH_BSN2H"],
    [1, 4096, 4096, 1, 512, "Q_K_V_BNSH"],
]


@pytest.mark.skipif(not ke.is_composable_kernel_available(), reason="ck is not enabled")
@pytest.mark.parametrize("mask_dim", [0], ids=get_mask_dim_id)
@pytest.mark.parametrize("biased", [False], ids=get_biased_id)
@pytest.mark.parametrize("causal", [False, True])
@pytest.mark.parametrize("batch, seqlen, total_seqlen, nhead, head_size, qkv_format_name", stabel_diffusion_configs)
@pytest.mark.parametrize("dtype", dtypes)
def test_gemm_softmax_gemm_permute_ck_sd(
    dtype, batch, seqlen, total_seqlen, nhead, head_size, biased, causal, mask_dim, qkv_format_name
):
    qkv_format = getattr(ke.qkv_format, qkv_format_name)
    f = getattr(ke, get_ck_binding_name(dtype, biased, mask_dim != 0))
    scale = 1.0 / np.sqrt(head_size)
    _test_gemm_softmax_gemm_permute(
        f, dtype, batch, seqlen, total_seqlen, nhead, head_size, biased, mask_dim, scale, causal, qkv_format
    )


@dataclass
class GemmSoftmaxGemmPermuteMetric(ke.ComputeMetric):
    batch: int
    seqlen: int
    total_seqlen: int
    num_heads: int
    head_size: int
    biased: bool
    mask_dim: int

    def report(self):
        bias_str = " biased" if self.biased else ""
        mask_str = f" mask_{self.mask_dim}d" if self.mask_dim != 0 else ""
        common = (
            f"{self.dtype} B={self.batch} S={self.seqlen} T={self.total_seqlen} "
            f"N={self.num_heads} H={self.head_size}{bias_str}{mask_str}, "
            f"{self.name}"
        )
        if self.duration <= 0:
            return "not supported          " + common

        return f"{self.duration:>6.2f} us {self.tflops:>5.2f} tflops " + common


@ke.dispatchable(pattern_arg=0)
def profile_gemm_softmax_gemm_permute_func(
    f, dtype, batch, seqlen, total_seqlen, num_heads, head_size, biased, mask_dim, scale, causal, qkv_format
):
    v_head_size = head_size
    q_shape = [batch, num_heads, seqlen, head_size]
    k_shape = [batch, num_heads, total_seqlen, head_size]
    v_shape = [batch, num_heads, total_seqlen, v_head_size]
    out_shape = [batch, seqlen, num_heads, head_size]

    attn_bias = None
    bias_shape = [batch, num_heads, seqlen, total_seqlen] if biased else None

    attn_mask = None
    mask_shape = None
    max_seqlen = None
    if mask_dim != 0:
        if mask_dim == 2:
            mask_shape = [batch, total_seqlen]
        elif mask_dim == 3:
            mask_shape = [batch, seqlen, total_seqlen]
        elif mask_dim == 4:
            max_seqlen = ((seqlen - 1) // 1024 + 1) * 1024  # round up to multiple of 1024
            mask_shape = [batch, 1, max_seqlen, max_seqlen]
        else:
            raise ValueError

    np.random.seed(42)
    q = multinormal_distribution(np.prod(q_shape[:-1]), q_shape[-1]).reshape(q_shape).astype(np.float64)
    k = multinormal_distribution(np.prod(k_shape[:-1]), k_shape[-1]).reshape(k_shape).astype(np.float64)
    v = multinormal_distribution(np.prod(v_shape[:-1]), v_shape[-1]).reshape(v_shape).astype(np.float64)
    if bias_shape is not None:
        attn_bias = np.random.uniform(-2, 2, size=bias_shape)
    if mask_shape is not None:
        attn_mask = (np.random.randint(0, 100, size=mask_shape) < 95).astype(np.int32)

    out = np.empty(out_shape, dtype=dtype)
    host_q, host_k, host_v = maybe_pack_q_k_v_bnsh_for_device_on_host(q, k, v, dtype, qkv_format)
    host_attn_bias = attn_bias.astype(dtype) if attn_bias is not None else None
    dev_q = ke.DeviceArray(host_q)
    dev_k = ke.DeviceArray(host_k) if host_k is not None else None
    dev_v = ke.DeviceArray(host_v) if host_v is not None else None
    dev_out = ke.DeviceArray(out)
    dev_attn_bias = ke.DeviceArray(host_attn_bias) if host_attn_bias is not None else None
    dev_attn_mask = ke.DeviceArray(attn_mask) if attn_mask is not None else None

    my_gemm_softmax_gemm_permute = f(
        batch,
        seqlen,
        total_seqlen,
        max_seqlen,
        num_heads,
        head_size,
        mask_dim,
        scale,
        causal,
        qkv_format,
        dev_q,
        dev_k,
        dev_v,
        dev_attn_bias,
        dev_attn_mask,
        dev_out,
    )

    for impl in my_gemm_softmax_gemm_permute.ListOps():
        duration_ms = -1
        if my_gemm_softmax_gemm_permute.SelectOp(impl):
            duration_ms = my_gemm_softmax_gemm_permute.Profile()

        m, n, k, o, gemm_batch = seqlen, total_seqlen, head_size, head_size, batch * num_heads
        flops_per_batch = m * n * k * 2 + m * n * o * 2
        flops_count_bias_and_softmax = True  # set to false to be aligned with ck
        if flops_count_bias_and_softmax:
            flops_per_batch += 2 * n + 1
        if flops_count_bias_and_softmax and attn_bias is not None:
            flops_per_batch += m * n
        if flops_count_bias_and_softmax and attn_mask is not None:
            flops_per_batch += m * n
        flops = flops_per_batch * gemm_batch

        ke.report(
            GemmSoftmaxGemmPermuteMetric(
                impl, dtype, duration_ms, flops, batch, seqlen, total_seqlen, num_heads, head_size, biased, mask_dim
            )
        )


@ke.dispatchable
def profile_with_args(
    dtype,
    batch,
    seqlen,
    total_seqlen,
    num_heads,
    head_size,
    biased,
    causal,
    mask_dim,
    scale,
    qkv_format,
):
    with ke.benchmark():
        args = (dtype, batch, seqlen, total_seqlen, num_heads, head_size, biased, mask_dim, scale, causal, qkv_format)
        if qkv_format == ke.qkv_format.Q_K_V_BNSH:
            profile_gemm_softmax_gemm_permute_func(
                getattr(ke, "GemmSoftmaxGemmPermuteGeneric_" + dtype_to_suffix(dtype)), *args
            )
        if ke.is_composable_kernel_available():
            profile_gemm_softmax_gemm_permute_func(
                getattr(ke, get_ck_binding_name(dtype, biased, mask_dim != 0)), *args
            )
        profile_gemm_softmax_gemm_permute_func(
            getattr(ke, "GemmSoftmaxGemmPermuteTunable_" + dtype_to_suffix(dtype)), *args
        )


def profile():
    for batch, seqlen, total_seqlen, nhead, head_size, qkv_format_name in stabel_diffusion_configs:
        profile_with_args(
            "float16",
            batch,
            seqlen,
            total_seqlen,
            nhead,
            head_size,
            biased=False,
            causal=False,
            mask_dim=0,
            qkv_format=getattr(ke.qkv_format, qkv_format_name),
            scale=0.125,
        )
        print()

    for args in product(dtypes, batches, seqlens, total_seqlens, num_heads, head_sizes, biaseds, causals, mask_dims):
        profile_with_args(*args, qkv_format=ke.qkv_format.Q_K_V_BNSH, scale=0.125)
        print()


if __name__ == "__main__":
    parser = ke.get_argument_parser()
    group = parser.add_argument_group()
    group.add_argument("dtype", choices=dtypes)
    group.add_argument("batch", type=int)
    group.add_argument("seqlen", type=int)
    group.add_argument("total_seqlen", type=int)
    group.add_argument("num_heads", type=int)
    group.add_argument("head_size", type=int)
    group.add_argument("biased", type=int, choices=[0, 1], default=0)
    group.add_argument("mask_dim", type=int, choices=[0, 2, 3, 4], default=2, help="0 for mask disabled")
    group.add_argument("causal", type=int, choices=[0, 1], default=0)
    group.add_argument("--scale", type=float, default=None, help="default to 1.0/sqrt(head_size)")
    group.add_argument(
        "--qkv_format",
        default="Q_K_V_BNSH",
        choices=[
            "Q_K_V_BNSH",  # non-packed, permuted
            "Q_K_V_BSNH",  # non-packed, non-permuted
            "Q_KV_BSNH_BSN2H",  # kv packed, non-permuted
            "QKV_BSN3H",  # qkv packed, non-permuted
        ],
    )

    if not ke.has_args():
        profile()
    else:
        args = parser.parse_args()
        args.dispatch(
            args.dtype,
            args.batch,
            args.seqlen,
            args.total_seqlen,
            args.num_heads,
            args.head_size,
            args.biased,
            args.causal,
            args.mask_dim,
            args.scale,
            getattr(ke.qkv_format, args.qkv_format),
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

from dataclasses import dataclass

import kernel_explorer as ke
import numpy as np
import pytest
from ml_dtypes import finfo, float8_e4m3fn, float8_e4m3fnuz
from utils import dtype_to_bytes, dtype_to_suffix, get_gemm_bert_sizes, matmul, transab_to_suffix


def create_device_array(a):
    ptr = a.__array_interface__["data"][0]
    size = a.size
    itemsize = finfo(a.dtype).bits // 8
    return ke.DeviceArray(ptr, size, itemsize)


def compute_scaling_factor(a: np.ndarray, fp8_max: float, margin: int) -> np.ndarray:
    amax = np.abs(a).max()
    scale = (fp8_max - margin) / amax  # fallback scale
    exp = np.floor(np.log2(fp8_max / amax)) - margin
    sf = np.round(np.power(2, np.abs(exp)))
    sf = np.where(amax > 0.0, sf, scale)
    sf = np.where(np.isfinite(amax), sf, scale)
    sf = np.where(exp < 0, 1 / sf, sf)

    return sf


def cast_and_scale(a, dtype: str):
    if dtype == "float16":
        return a.astype(dtype), 1.0
    elif np.dtype(dtype) in (float8_e4m3fn, float8_e4m3fnuz):
        t = globals()[dtype]
        sf = compute_scaling_factor(a, fp8_max=finfo(t).max, margin=4)
        return (a * sf).astype(t), sf
    else:
        raise ValueError(dtype)


@ke.dispatchable(pattern_arg=0)
def _test_gemm(
    func, dta: str, dtb: str, dtc: str, transa: bool, transb: bool, m: int, n: int, k: int, alpha=1.0, beta=0.0
):
    assert beta == 0.0, "beta is not supported"
    assert dta in ["float16", "float8_e4m3fn", "float8_e4m3fnuz"]
    assert dtb in ["float16", "float8_e4m3fn", "float8_e4m3fnuz"]
    assert dtc in ["float16"]

    a_shape = (k, m) if transa else (m, k)
    b_shape = (n, k) if transb else (k, n)

    np.random.seed(0)

    a, scale_a = cast_and_scale(np.random.rand(*a_shape), dta)
    b, scale_b = cast_and_scale(np.random.rand(*b_shape), dtb)
    scale_c = float("nan")

    inv_scale_a = np.array(1 / scale_a).astype("float32")
    inv_scale_b = np.array(1 / scale_b).astype("float32")
    inv_scale_c = np.array(1 / scale_c).astype("float32")

    ref_c = matmul(a * inv_scale_a, b * inv_scale_b, transa, transb)
    if alpha != 1.0:
        ref_c *= alpha

    my_c = np.ones((m, n), dtype=dtc)
    dev_a = create_device_array(a)
    dev_b = create_device_array(b)
    dev_c = create_device_array(my_c)
    dev_inv_scale_a = create_device_array(inv_scale_a)
    dev_inv_scale_b = create_device_array(inv_scale_b)
    dev_inv_scale_c = create_device_array(inv_scale_c)

    opa = ke.blas_op.T if transa else ke.blas_op.N
    opb = ke.blas_op.T if transb else ke.blas_op.N
    lda = a_shape[1]
    ldb = b_shape[1]
    my_gemm = func(
        opa,
        opb,
        m,
        n,
        k,
        alpha,
        dev_a,
        lda,
        dev_inv_scale_a,
        dev_b,
        ldb,
        dev_inv_scale_b,
        beta,
        dev_c,
        n,
        dev_inv_scale_c,
    )

    failures = {}

    # TODO: how to derive the bound for fp8?
    atol = 0.01
    rtol = 0.005
    print(f"atol={atol} rtol={rtol}")  # print for pytest -s -v

    for impl in my_gemm.ListOps():
        if not my_gemm.SelectOp(impl):
            continue
        # Restore C Array
        my_c.fill(1.0)
        dev_c.UpdateDeviceArray()
        my_gemm.Run()
        dev_c.UpdateHostNumpyArray()

        try:
            np.testing.assert_allclose(my_c, ref_c, atol=atol, rtol=rtol)
        except Exception as err:
            header = "*" * 30 + impl + "*" * 30
            print(header)
            print(err)
            print("*" * len(header))
            failures[impl] = str(err)

    if failures:
        raise Exception(failures)


dtypes = [
    ("float8_e4m3fn", "float16", "float16"),
    ("float8_e4m3fnuz", "float16", "float16"),
    ("float16", "float8_e4m3fn", "float16"),
    ("float16", "float8_e4m3fnuz", "float16"),
]
all_transabs = [(False, False), (False, True)]


@pytest.mark.skipif(not ke.is_float8_available(), reason="float8 is not enabled")
@pytest.mark.skipif(not ke.is_composable_kernel_available(), reason="ck is not enabled")
@pytest.mark.parametrize(
    "m, n, k",
    [
        (1, 768, 768),
        (768, 768, 768),
        (1, 8192, 28672),
        (1, 28672, 8192),
        (1, 8192, 8192),
        (128, 8192, 28672),
        (128, 28672, 8192),
        (128, 8192, 8192),
    ],
)
@pytest.mark.parametrize("transa, transb", all_transabs)
@pytest.mark.parametrize("dta, dtb, dtc", dtypes)
@ke.dispatchable
def test_ck_gemm(dta, dtb, dtc, transa, transb, m, n, k):
    if dtb == "float16" and transb:
        pytest.skip("Only supports transb when b is fp8")
    wrapper_name = f"GemmFloat8CK_{dtype_to_suffix(dta)}_{dtype_to_suffix(dtb)}_{dtype_to_suffix(dtc)}_{transab_to_suffix((transa, transb))}"
    _test_gemm(getattr(ke, wrapper_name), dta, dtb, dtc, transa, transb, m, n, k)


@pytest.mark.skipif(not ke.is_float8_available(), reason="float8 is not enabled")
@pytest.mark.skipif(not ke.is_composable_kernel_available(), reason="ck is not enabled")
@pytest.mark.parametrize("alpha, beta", [(1.5, 0.0), [2.0, 0.0]])
@pytest.mark.parametrize("m, n, k", [(768, 768, 768)])
@pytest.mark.parametrize("transa, transb", all_transabs)
@pytest.mark.parametrize("dta, dtb, dtc", dtypes)
def test_ck_gemm_alpha_beta(dta, dtb, dtc, transa, transb, m, n, k, alpha, beta):
    if dtb == "float16" and transb:
        pytest.skip("Only supports transb when b is fp8")
    wrapper_name = f"GemmFloat8CK_{dtype_to_suffix(dta)}_{dtype_to_suffix(dtb)}_{dtype_to_suffix(dtc)}_{transab_to_suffix((transa, transb))}"
    _test_gemm(getattr(ke, wrapper_name), dta, dtb, dtc, transa, transb, m, n, k, alpha, beta)


@pytest.mark.skipif(not ke.is_float8_available(), reason="float8 is not enabled")
@pytest.mark.skipif(not ke.is_composable_kernel_available(), reason="ck is not enabled")
@pytest.mark.parametrize("alpha, beta", [(1.5, 0.0), [2.0, 0.0]])
@pytest.mark.parametrize("m, n, k", [(256, 256, 256)])
@pytest.mark.parametrize("transa, transb", all_transabs)
@pytest.mark.parametrize("dta, dtb, dtc", dtypes)
def test_tunable_gemm(dta, dtb, dtc, transa, transb, m, n, k, alpha, beta):
    if dtb == "float16" and transb:
        pytest.skip("Only supports transb when b is fp8")
    wrapper_name = f"GemmFloat8Tunable_{dtype_to_suffix(dta)}_{dtype_to_suffix(dtb)}_{dtype_to_suffix(dtc)}_{transab_to_suffix((transa, transb))}"
    _test_gemm(getattr(ke, wrapper_name), dta, dtb, dtc, transa, transb, m, n, k, alpha, beta)


@dataclass
class GemmMetric(ke.BandwidthMetric, ke.ComputeMetric):
    transa: bool
    transb: bool
    m: int
    n: int
    k: int

    def report(self):
        common = (
            f"{self.dtype} {transab_to_suffix((self.transa, self.transb))} "
            f"m={self.m:<4} n={self.n:<4} k={self.k:<4} {self.name}"
        )
        if self.duration <= 0:
            return "not supported          " + common

        return f"{self.duration:>6.2f} us {self.tflops:>5.2f} tflops {self.gbps:5.2f} GB/s " + common


@ke.dispatchable(pattern_arg=0)
def profile_gemm_func(
    func, dta: str, dtb: str, dtc: str, transa: bool, transb: bool, m: int, n: int, k: int, alpha=1.0, beta=0.0
):
    assert beta == 0.0, "beta is not supported"
    a_shape = (k, m) if transa else (m, k)
    b_shape = (n, k) if transb else (k, n)

    np.random.seed(0)
    a, scale_a = cast_and_scale(np.random.rand(*a_shape) + 0.1, dta)
    b, scale_b = cast_and_scale(np.random.rand(*b_shape) + 0.1, dtb)
    scale_c = 1.0

    inv_scale_a = np.array(1 / scale_a).astype("float32")
    inv_scale_b = np.array(1 / scale_b).astype("float32")
    inv_scale_c = np.array(1 / scale_c).astype("float32")

    my_c = np.ones((m, n), dtype=dtc)

    dev_a = create_device_array(a)
    dev_b = create_device_array(b)
    dev_c = create_device_array(my_c)
    dev_inv_scale_a = create_device_array(inv_scale_a)
    dev_inv_scale_b = create_device_array(inv_scale_b)
    dev_inv_scale_c = create_device_array(inv_scale_c)

    opa = ke.blas_op.T if transa else ke.blas_op.N
    opb = ke.blas_op.T if transb else ke.blas_op.N
    lda = a_shape[1]
    ldb = b_shape[1]
    my_gemm = func(
        opa,
        opb,
        m,
        n,
        k,
        alpha,
        dev_a,
        lda,
        dev_inv_scale_a,
        dev_b,
        ldb,
        dev_inv_scale_b,
        beta,
        dev_c,
        n,
        dev_inv_scale_c,
    )

    for impl in my_gemm.ListOps():
        duration_ms = -1
        if my_gemm.SelectOp(impl):
            duration_ms = my_gemm.Profile()
        FLOPs = m * k * n * 2  # noqa: N806
        total_bytes = m * k * dtype_to_bytes(dta) + k * n * dtype_to_bytes(dtb) + m * n * dtype_to_bytes(dtc)

        ke.report(GemmMetric(impl, f"{dta}_{dtb}_{dtc}", duration_ms, FLOPs, total_bytes, transa, transb, m, n, k))


@ke.dispatchable
def profile_with_args(dta, dtb, dtc, transa, transb, m, n, k):
    dtype_suffix = "_" + dtype_to_suffix(dta) + "_" + dtype_to_suffix(dtb) + "_" + dtype_to_suffix(dtc)
    transab_suffix = "_" + transab_to_suffix((transa, transb))
    with ke.benchmark():
        profile_gemm_func(
            getattr(ke, "GemmFloat8CK" + dtype_suffix + transab_suffix), dta, dtb, dtc, transa, transb, m, n, k
        )
        profile_gemm_func(
            getattr(ke, "GemmFloat8Tunable" + dtype_suffix + transab_suffix), dta, dtb, dtc, transa, transb, m, n, k
        )
    print()


def profile():
    for dta, dtb, dtc in dtypes:
        for m, n, k in get_gemm_bert_sizes(full=True):
            profile_with_args(dta, dtb, dtc, False, False, m, n, k)


if __name__ == "__main__":
    parser = ke.get_argument_parser()
    group = parser.add_argument_group()
    group.add_argument("dta", choices=["float8_e4m3fn", "float8_e4m3fnuz", "float16"])
    group.add_argument("dtb", choices=["float8_e4m3fn", "float8_e4m3fnuz", "float16"])
    group.add_argument("dtc", choices=["float8_e4m3fn", "float8_e4m3fnuz", "float16"])
    group.add_argument("transa", choices="NT")
    group.add_argument("transb", choices="NT")
    group.add_argument("m", type=int)
    group.add_argument("n", type=int)
    group.add_argument("k", type=int)

    if not ke.has_args():
        profile()
    else:
        args = parser.parse_args()
        args.dispatch(args.dta, args.dtb, args.dtc, args.transa == "T", args.transb == "T", args.m, args.n, args.k)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import logging

from fusion_attention_sam2 import FusionMultiHeadAttentionSam2
from fusion_layernorm import FusionLayerNormalizationNCHW
from fusion_options import FusionOptions
from import_utils import is_installed
from onnx import ModelProto
from onnx_model_bert import BertOnnxModel

logger = logging.getLogger(__name__)


class Sam2OnnxModel(BertOnnxModel):
    def __init__(self, model: ModelProto, num_heads: int = 0, hidden_size: int = 0):
        """Initialize SAM2 ONNX Model.

        Args:
            model (ModelProto): the ONNX model
            num_heads (int, optional): number of attention heads. Defaults to 0 (detect the parameter automatically).
            hidden_size (int, optional): hidden dimension. Defaults to 0 (detect the parameter automatically).
        """
        assert (num_heads == 0 and hidden_size == 0) or (num_heads > 0 and hidden_size % num_heads == 0)

        super().__init__(model, num_heads=num_heads, hidden_size=hidden_size)

    def postprocess(self):
        self.prune_graph()
        self.remove_unused_constant()

    def fuse_layer_norm(self):
        super().fuse_layer_norm()

        fusion = FusionLayerNormalizationNCHW(self)
        fusion.apply()

    def fuse_multi_head_attention(self, options: FusionOptions | None = None):
        mha_fusion = FusionMultiHeadAttentionSam2(self, self.hidden_size, self.num_heads)
        mha_fusion.apply()

    def optimize(self, options: FusionOptions | None = None, add_dynamic_axes: bool = False):
        if is_installed("tqdm"):
            import tqdm
            from tqdm.contrib.logging import logging_redirect_tqdm

            with logging_redirect_tqdm():
                steps = 12
                progress_bar = tqdm.tqdm(range(steps), initial=0, desc="sam2 fusion")
                self._optimize(options, progress_bar)
        else:
            logger.info("tqdm is not installed. Run optimization without progress bar")
            self._optimize(options, None)

    def _optimize(self, options: FusionOptions | None = None, progress_bar=None):
        if (options is not None) and not options.enable_shape_inference:
            self.disable_shape_inference()

        self.utils.remove_identity_nodes()
        if progress_bar:
            progress_bar.update(1)

        # Remove cast nodes that having same data type of input and output based on symbolic shape inference.
        self.utils.remove_useless_cast_nodes()
        if progress_bar:
            progress_bar.update(1)

        if (options is None) or options.enable_layer_norm:
            self.fuse_layer_norm()
        if progress_bar:
            progress_bar.update(1)

        if (options is None) or options.enable_gelu:
            self.fuse_gelu()
        if progress_bar:
            progress_bar.update(1)

        self.fuse_reshape()
        if progress_bar:
            progress_bar.update(1)

        if (options is None) or options.enable_attention:
            self.fuse_multi_head_attention(options)
        if progress_bar:
            progress_bar.update(1)

        if (options is None) or options.enable_skip_layer_norm:
            self.fuse_skip_layer_norm()
        if progress_bar:
            progress_bar.update(1)

        self.fuse_shape()
        if progress_bar:
            progress_bar.update(1)

        # Remove reshape nodes that having same shape of input and output based on symbolic shape inference.
        self.utils.remove_useless_reshape_nodes()
        if progress_bar:
            progress_bar.update(1)

        if (options is None) or options.enable_bias_skip_layer_norm:
            # Fuse SkipLayerNormalization and Add Bias before it.
            self.fuse_add_bias_skip_layer_norm()
        if progress_bar:
            progress_bar.update(1)

        if options is not None and options.enable_gelu_approximation:
            self.gelu_approximation()
        if progress_bar:
            progress_bar.update(1)

        self.postprocess()
        if progress_bar:
            progress_bar.update(1)

        logger.info(f"opset version: {self.get_opset_version()}")

    def get_fused_operator_statistics(self):
        """
        Returns node count of fused operators.
        """
        op_count = {}
        ops = [
            "MultiHeadAttention",
            "LayerNormalization",
            "SkipLayerNormalization",
        ]

        for op in ops:
            nodes = self.get_nodes_by_op_type(op)
            op_count[op] = len(nodes)

        logger.info(f"Optimized operators:{op_count}")
        return op_count



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# sampler.py

import math
from collections.abc import Callable, Iterator

import numpy as np
import torch
import torch.distributed as dist
from torch.utils.data.dataset import Dataset
from torch.utils.data.sampler import Sampler


def _shard_wrapped_indices_across_workers(dataset_index_list, num_shards, num_samples_per_shard):
    """Yield successive num_shards-sized chunks from dataset_index_list."""
    num_samples = max(1, num_samples_per_shard)
    num_elements = num_samples * num_shards
    current_lst = []
    for i in range(num_elements):
        current_lst.append(dataset_index_list[i % len(dataset_index_list)])
        if len(current_lst) == num_shards:
            yield current_lst
            current_lst = []


def shard_wrapped_indices_for_worker(dataset_index_list, shard_id, num_shards):
    """Shard wrapped around dataset_index_list across num_shards and return the indices for this shard_id"""
    num_samples_per_worker = (len(dataset_index_list) + num_shards - 1) // num_shards
    sharded_indices = list(
        _shard_wrapped_indices_across_workers(dataset_index_list, num_shards, num_samples_per_worker)
    )
    return [sharded_indices[i][shard_id] for i in range(len(sharded_indices))]


# Implementation is adapted from bagua/load_balancing_data_loader.py
# https://github.com/BaguaSys/bagua/blob/01874a7c3f90904c37c5612a9db866b5d4b8b5ed/bagua/torch_api/contrib/load_balancing_data_loader.py#L12
class LoadBalancingDistributedSampler:
    r"""Sampler that balances the data load across workers based on the sample's complexity.
    This sampler uses a :attr:`complexity_fn` to calculate each sample's computational
    complexity and make each batch get similar computational complexity.
    This is useful in scenarios like speech and NLP, where each batch has variable
    length and distributed training suffers from straggler problem. In such scenarios,
    the complexity function could be defined to return the length of the input sample sequence.
    The usage is similar to `torch.utils.data.DistributedSampler`, where each process loads a
    subset of the original dataset that is exclusive to it.
    The sampler sorts the dataset in increasing order of complexity. If the :attr:`group_size` is
    provided, the sorting happens within dataset groups of size :attr:`group_size` before the
    group order is shuffled followed by sharding of data across workers. If :attr:`group_size`
    is not provided, the data is distributed across workers before the data indices for each worker
    is shuffled deterministically.
    .. note::
        Dataset is assumed to be of constant size (map-style dataset).
    Args:
        dataset: Dataset (map-style) used for sampling.
        complexity_fn(Callable): A function whose input is a sample and output is an integer as a
            measure of the computational complexity of the sample.
        world_size (int, optional): Number of processes participating in
            distributed training. By default, :attr:`world_size` is retrieved from the
            current distributed group.
        rank (int, optional): Rank of the current process within :attr:`world_size`.
            By default, :attr:`rank` is retrieved from the current distributed
            group.
        shuffle (bool, optional): If ``True`` (default), sampler will shuffle the
            indices within the dataset if :attr:`group_size` is None, else will
            shuffle the groups if :attr:`group_size` is not None.
        group_size (int, optional): If provided, the dataset will be broken down into
            :attr:`group_size` sized groups. Indices will only be sorted within the groups
            and not across the entire dataset. If :attr:`shuffle` is ```True``` and
            :attr:`group_size` is not ```None```, the position of each group in the dataset
            will be shuffled. Default: ```None```
        seed (int, optional): random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Default: 0.
        drop_last (bool, optional): if ``True``, then the sampler will drop the
            tail of the data to make it evenly divisible across the number of
            shards. If ``False``, the sampler will add extra indices to make
            the data evenly divisible across the shards. Default: ``False``.
        random_level (float, optional): A float varies from 0 and 1 that controls the extent
            of load balance. 0 means the best load balance, while 1 means the opposite.
    .. warning::
        In distributed mode, calling the :meth:`set_epoch` method at
        the beginning of each epoch **before** creating the `torch.utils.data.DataLoader` iterator
        is necessary to make shuffling work properly across multiple epochs. Otherwise,
        the same ordering will be always used.
    Example::
        Define your :attr:`complexity_fn`, which accepts a dataset sample as its input and produces an integer
        as the sample's computational complexity:
        >>> dataset = MyVariableSequenceLengthDataset(dataset_samples)
        >>> complexity_fn = lambda x: len(x)
        Below is the usage of :class:`LoadBalancingDistributedSampler`
        and `torch.utils.data.DataLoader`:
        >>> sampler = onnxruntime.training.utils.data.LoadBalancingDistributedSampler(
        ...     dataset,
        ...     complexity_fn=complexity_fn)
        >>> loader = torch.utils.data.DataLoader(dataset,
        ...     sampler=sampler)
        >>>
        >>> for epoch in range(start_epoch, n_epochs):
        ...     if is_distributed:
        ...         sampler.set_epoch(epoch)
        ...     train(loader)
    """

    def __init__(
        self,
        dataset: Dataset,
        complexity_fn: Callable[..., int],
        world_size: int | None = None,
        rank: int | None = None,
        shuffle: bool = True,
        group_size: int | None = None,
        seed: int = 0,
        drop_last: bool = False,
        random_level: float = 0,
    ) -> None:
        if world_size is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            world_size = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        if rank >= world_size or rank < 0:
            raise ValueError(f"Invalid rank {rank}, rank should be in the interval [0, {world_size - 1}]")
        self.dataset = dataset
        self.world_size = world_size
        self.rank = rank
        self.epoch = 0
        self.drop_last = drop_last
        self.group_size = group_size

        # If the dataset length is evenly divisible by number of shards, then there
        # is no need to drop any data, since the dataset will be split equally.
        dataset_len = len(self.dataset)
        if self.drop_last and dataset_len % self.world_size != 0:
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data when
            # using this Sampler.
            self.num_samples = dataset_len // self.world_size
        else:
            self.num_samples = math.ceil(dataset_len / self.world_size)
        self.total_size = self.num_samples * self.world_size
        self.shuffle = shuffle
        self.seed = seed

        self.complexity_fn = complexity_fn
        self.sample_complexities = None
        self.ordered_sample_complexities = None

        if random_level < 0.0 or random_level > 1.0:
            raise ValueError(f"Invalid random level {random_level}, shoule be in the range [0.0, 1.0]")

        self.random_level = random_level
        self.random_number = None

    def _sort_shard_and_shuffle_dataset(self):
        # This method returns a list of dataset sample indices after
        # the dataset has been sorted, sharded and shuffled.
        # The sorting of the dataset happens based on the group_size and complexities
        # of each sample.
        # Sharding happens across the number of workers.
        # Shuffling is done either before sharding on the group indices (if group_size is provided)
        # or on the dataset sample indices if the group_size is not provided.

        def sort_in_groups(sample_complexities, group_size):
            """Sort the dataset samples indices inside each group of size group_size."""
            # If the group_size is None, the entire dataset is considered as a single group
            if group_size is None:
                group_size = len(sample_complexities)
            # Sort the dataset samples inside each group of the dataset based on sample complexity.
            for group_begin_index in range(0, len(sample_complexities), group_size):
                group_end_index = min(group_begin_index + group_size, len(sample_complexities))
                sorted_indices = group_begin_index + np.argsort(
                    sample_complexities[group_begin_index:group_end_index, 1]
                )
                sample_complexities[group_begin_index:group_end_index, :] = sample_complexities[sorted_indices]
            return sample_complexities

        # Get the samples and their complexities from the complexity_fn
        if not self.sample_complexities:
            self.sample_complexities = np.empty((len(self.dataset), 2), dtype=np.int64)
            for sample_index in range(len(self.dataset)):
                self.sample_complexities[sample_index][0] = sample_index
                self.sample_complexities[sample_index][1] = self.complexity_fn(self.dataset[sample_index])

        if self.random_number is None:
            max_complexity = max(self.sample_complexities, key=lambda t: t[1])[1]
            min_complexity = min(self.sample_complexities, key=lambda t: t[1])[1]
            self.random_number = int((max_complexity - min_complexity) * self.random_level + 1)

        sample_complexities = self.sample_complexities.copy()

        # Control the degree of load balancing by modifying the complexities of
        # all samples using the random_number.
        g = torch.Generator()
        g = g.manual_seed(self.seed + self.epoch)

        if self.random_number > 1:
            complexity_random_ints = torch.randint(
                self.random_number, (len(sample_complexities),), generator=g
            ).tolist()

            for index, random_int in enumerate(complexity_random_ints):
                sample_complexities[index][1] += random_int

        # Sort the data based on the computed complexities and group sizes.
        # Sort only once if random_number <= 1 else sort everytime
        if self.ordered_sample_complexities is None or self.random_number > 1:
            self.ordered_sample_complexities = sort_in_groups(sample_complexities, self.group_size)
        ordered_sample_complexities = self.ordered_sample_complexities

        # If group_size is not None, shuffle the index of each group instead
        # of shuffling the data indices.
        if self.shuffle and self.group_size is not None:
            num_groups = (len(self.sample_complexities) + self.group_size - 1) // self.group_size
            group_order = torch.randperm(num_groups, generator=g).tolist()
            end = 0
            sample_complexities_copy = ordered_sample_complexities.copy()
            for group_index in group_order:
                original_list_begin_index = self.group_size * group_index
                original_list_end_index = min(original_list_begin_index + self.group_size, len(sample_complexities))
                begin = end
                end = begin + (original_list_end_index - original_list_begin_index)
                sample_complexities_copy[begin:end, :] = sample_complexities[
                    original_list_begin_index:original_list_end_index, :
                ]
            ordered_sample_complexities = sample_complexities_copy

        # Shard the data across the different workers.
        index_chunks = list(
            _shard_wrapped_indices_across_workers(
                [index_complexity_tuple[0] for index_complexity_tuple in ordered_sample_complexities],
                self.world_size,
                self.num_samples,
            )
        )

        # Shuffle the sharded data indices deterministically based on epoch and seed.
        chunk_indices = list(range(len(index_chunks)))
        if self.shuffle and self.group_size is None:
            chunk_indices = torch.randperm(len(index_chunks), generator=g).tolist()

        if not self.drop_last:
            # Add extra samples to make it evenly divisible
            padding_size = self.num_samples - len(chunk_indices)
            if padding_size <= len(chunk_indices):
                chunk_indices += chunk_indices[:padding_size]
            else:
                chunk_indices += (chunk_indices * math.ceil(padding_size / len(chunk_indices)))[:padding_size]
        else:
            # Remove tail of data to make it evenly divisible.
            chunk_indices = chunk_indices[: self.num_samples]

        assert len(chunk_indices) == self.num_samples
        return index_chunks, chunk_indices

    def __iter__(self) -> Iterator:
        index_chunks, chunk_indices = self._sort_shard_and_shuffle_dataset()
        # Extract indices based on current rank.
        indices = [index_chunks[i][self.rank] for i in chunk_indices]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self) -> int:
        return self.num_samples

    def set_epoch(self, epoch: int) -> None:
        r"""Sets the epoch for this sampler.
        When :attr:`shuffle=True`, this ensures all shards use a different
        random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.
        Args:
            epoch (int): Epoch number.
        """
        self.epoch = epoch


class LoadBalancingDistributedBatchSampler(Sampler):
    r"""Wraps another load balance sampler to yield variable sized mini-batches.
    Args:
        sampler (LoadBalancingDistributedSampler): Load balance sampler.
        batch_fn (Callable): Callable to yield mini-batch indices.
        drop_last (bool): If ``True``, the sampler will drop the last few batches exceeding
            the least number of batches among replicas, otherwise, the number of batches
            on each replica will be padded to the same.
    :attr:`batch_fn` will have the signature of::
        def batch_fn(indices: List[int]) -> List[List[int]]
    Example::
        >>> from onnxruntime.training.utils.data import LoadBalancingDistributedSampler, \
        ...     LoadBalancingDistributedBatchSampler
        >>>
        >>> sampler = LoadBalancingDistributedSampler(dataset, complexity_fn=complexity_fn)
        >>> batch_sampler = LoadBalancingDistributedBatchSampler(sampler, batch_fn=batch_fn)
        >>> loader = torch.utils.data.DataLoader(dataset, batch_sampler=batch_sampler)
        >>>
        >>> for epoch in range(start_epoch, n_epochs):
        ...     batch_sampler.set_epoch(epoch)
        ...     train(loader)
    """

    def __init__(
        self,
        sampler: LoadBalancingDistributedSampler,
        batch_fn,
        drop_last: bool = False,
    ) -> None:
        if not isinstance(sampler, LoadBalancingDistributedSampler):
            raise ValueError("sampler should be of LoadBalancingDistributedSampler type.")

        if sampler.drop_last:
            raise ValueError("drop_last of sampler should be False")

        self.sampler = sampler
        self.batch_fn = batch_fn
        self.drop_last = drop_last

        self.world_size = self.sampler.world_size
        self.rank = self.sampler.rank

        self.generate_batches()

    def generate_batches(self):
        index_chunks, chunk_indices = self.sampler._sort_shard_and_shuffle_dataset()

        batches = []
        for rank in range(self.world_size):
            sub_indices = [index_chunks[i][rank] for i in chunk_indices]
            batches.append(self.batch_fn(sub_indices))

        self.total_batch = max([len(b) for b in batches]) if not self.drop_last else min([len(b) for b in batches])

        # here {len(batches[self.rank]) - self.total_batch} batches dropped for
        # rank {self.rank}
        if self.total_batch < len(batches[self.rank]):
            pass

        self.padded_batches = [batch + batch[: self.total_batch - len(batch)] for batch in batches]

    def __iter__(self):
        return iter(self.padded_batches[self.rank])

    def __len__(self):
        return self.total_batch

    def set_epoch(self, epoch: int) -> None:
        r"""
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.
        Args:
            epoch (int): Epoch number.
        """
        self.sampler.set_epoch(epoch)
        self.generate_batches()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# deprecated: no longer using kernel def hashes
class DeprecatedSessionState(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DeprecatedSessionState()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsDeprecatedSessionState(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def DeprecatedSessionStateBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x4F\x52\x54\x4D", size_prefixed=size_prefixed)

    # DeprecatedSessionState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DeprecatedSessionState
    def Kernels(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from ort_flatbuffers_py.fbs.DeprecatedKernelCreateInfos import DeprecatedKernelCreateInfos
            obj = DeprecatedKernelCreateInfos()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DeprecatedSessionState
    def SubGraphSessionStates(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from ort_flatbuffers_py.fbs.DeprecatedSubGraphSessionState import DeprecatedSubGraphSessionState
            obj = DeprecatedSubGraphSessionState()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # DeprecatedSessionState
    def SubGraphSessionStatesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # DeprecatedSessionState
    def SubGraphSessionStatesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def DeprecatedSessionStateStart(builder):
    builder.StartObject(2)

def Start(builder):
    DeprecatedSessionStateStart(builder)

def DeprecatedSessionStateAddKernels(builder, kernels):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(kernels), 0)

def AddKernels(builder, kernels):
    DeprecatedSessionStateAddKernels(builder, kernels)

def DeprecatedSessionStateAddSubGraphSessionStates(builder, subGraphSessionStates):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(subGraphSessionStates), 0)

def AddSubGraphSessionStates(builder, subGraphSessionStates):
    DeprecatedSessionStateAddSubGraphSessionStates(builder, subGraphSessionStates)

def DeprecatedSessionStateStartSubGraphSessionStatesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartSubGraphSessionStatesVector(builder, numElems: int) -> int:
    return DeprecatedSessionStateStartSubGraphSessionStatesVector(builder, numElems)

def DeprecatedSessionStateEnd(builder):
    return builder.EndObject()

def End(builder):
    return DeprecatedSessionStateEnd(builder)



"""
@brief      test log(time=3s)
"""

import copy
import unittest

import numpy as np
import torch

from onnxruntime.training.ortmodule import ORTModule


class TestOnnxOpsOrtModule(unittest.TestCase):
    def assert_values_are_close(self, tensor, other, rtol=1e-05, atol=1e-06):
        are_close = torch.allclose(tensor, other, rtol=rtol, atol=atol)
        if not are_close:
            abs_diff = torch.abs(tensor - other)
            abs_other = torch.abs(other)
            max_atol = torch.max(abs_diff - rtol * abs_other)
            max_rtol = torch.max((abs_diff - atol) / abs_other)
            raise AssertionError(f"The maximum atol is {max_atol!r}, maximum rtol is {max_rtol!r}.")

    def assert_gradients_match_and_reset_gradient(
        self, ort_model, pt_model, none_pt_params=None, reset_gradient=True, rtol=1e-05, atol=1e-06
    ):
        if none_pt_params is None:
            none_pt_params = []
        ort_named_params = list(ort_model.named_parameters())
        pt_named_params = list(pt_model.named_parameters())
        self.assertEqual(len(ort_named_params), len(pt_named_params))

        for ort_named_param, pt_named_param in zip(ort_named_params, pt_named_params, strict=False):
            ort_name, ort_param = ort_named_param
            pt_name, pt_param = pt_named_param

            self.assertIn(pt_name, ort_name)
            if pt_name in none_pt_params:
                self.assertNotEmpty(pt_param.grad)
                if ort_param is not None:
                    self.assertFalse(torch.is_nonzero(torch.count_nonzero(ort_param.grad)))
            else:
                self.assert_values_are_close(ort_param.grad, pt_param.grad, rtol=rtol, atol=atol)

            if reset_gradient:
                ort_param.grad = None
                pt_param.grad = None

    def gradient_correctness(self, name, device, debug=False):
        pt_model_cls, op_grad_type, kwargs = self.get_torch_model_name(name, device)
        if kwargs is None:
            kwargs = {}
        N = 32  # noqa: N806
        pt_model = pt_model_cls().to(device)
        D_in = pt_model.fc1.in_features  # noqa: N806
        ort_model = ORTModule(copy.deepcopy(pt_model))

        def run_step(model, x):
            prediction = model(x)
            loss = prediction.sum()
            loss.backward()
            return prediction

        for _ in range(10):
            x = torch.randn(N, D_in, device=device)
            pt_prediction = run_step(pt_model, x)
            ort_prediction = run_step(ort_model, x)

            self.assert_values_are_close(ort_prediction, pt_prediction, **kwargs)
            self.assert_gradients_match_and_reset_gradient(ort_model, pt_model, **kwargs)

        onnx_graph_inf = ort_model._torch_module._execution_manager._training_manager._graph_transition_manager._exported_model_info.exported_model
        onnx_graph_train = ort_model._torch_module._execution_manager._training_manager._onnx_models.optimized_model
        if debug:
            with open(f"debug_{name}_ortmodule_infer.onnx", "wb") as f:
                f.write(onnx_graph_inf.SerializeToString())
            with open(f"debug_{name}_ortmodule_train.onnx", "wb") as f:
                f.write(onnx_graph_train.SerializeToString())
        self.assertIn(f'op_type: "{name}"', str(onnx_graph_inf))
        for onnx_model in [onnx_graph_inf, onnx_graph_train]:
            for oimp in onnx_model.opset_import:
                if oimp.domain == "":
                    self.assertEqual(oimp.version, 17)  # Needs to match latest default ORTModule opset
        if op_grad_type is not None:
            if isinstance(op_grad_type, tuple):
                text = str(onnx_graph_train)
                if all((f'op_type: "{op}"') not in text for op in op_grad_type):
                    raise AssertionError("Operator {} not found in {}.".format(" or ".join(op_grad_type), text))
            else:
                self.assertIn(f'op_type: "{op_grad_type}"', str(onnx_graph_train))

    def get_torch_model_name(self, name, device):
        def from_numpy(v, device=None, requires_grad=False):
            v = torch.from_numpy(v)
            if device is not None:
                v = v.to(device)
            v.requires_grad_(requires_grad)
            return v

        if name == "Softmax":

            class TestSoftmax(torch.nn.Module):
                def __init__(self, input_size=128, hidden_size=500, num_classes=100):
                    torch.nn.Module.__init__(self)
                    self.fc1 = torch.nn.Linear(input_size, hidden_size)
                    self.thfct = torch.nn.Softmax()
                    self.fc2 = torch.nn.Linear(hidden_size, num_classes)

                def forward(self, input1):
                    out = self.fc1(input1)
                    out = self.thfct(out)
                    out = self.fc2(out)
                    return out

            return TestSoftmax, ("SoftmaxGrad", "SoftmaxGrad_13"), None

        if name == "GatherElements":

            class TestGatherElement(torch.nn.Module):
                def __init__(self, input_size=32, hidden_size=500, num_classes=100):
                    torch.nn.Module.__init__(self)
                    self.fc1 = torch.nn.Linear(input_size, hidden_size)
                    rev_idx = np.array(list(np.arange(hidden_size)[::-1]), dtype=np.int64)
                    idx = np.empty((input_size, hidden_size), dtype=np.int64)
                    for i in range(idx.shape[0]):
                        idx[i, :] = rev_idx
                    self.indices = from_numpy(idx, device=device)
                    self.fc2 = torch.nn.Linear(hidden_size, num_classes)

                def forward(self, input1):
                    out = self.fc1(input1)
                    out = torch.gather(out, 1, self.indices)
                    out = self.fc2(out)
                    return out

            return TestGatherElement, "GatherElementsGrad", {"rtol": 1e-04, "atol": 1e-05}

        raise AssertionError(f"Unexpected name={name!r}.")

    def test_onnx_ops(self):
        for name in ["GatherElements", "Softmax"]:
            for device_name in ["cuda:0", "cpu"]:
                if device_name == "cuda:0" and not torch.cuda.is_available():
                    continue
                with self.subTest(name=name, device=device_name):
                    device = torch.device(device_name)
                    self.gradient_correctness(name, device)

    @unittest.skipIf(not torch.cuda.is_bf16_supported(), "Test requires CUDA and BF16 support")
    def test_softmax_bf16_large(self):
        raise unittest.SkipTest("Temporarily disabled pending investigation")

        if torch.version.cuda is None:
            # Only run this test when CUDA is available, as on ROCm BF16 is not supported by MIOpen.
            return

        class Model(torch.nn.Module):
            def __init__(self):
                super().__init__()

            def forward(self, input):
                out = torch.softmax(input, dim=-1)
                return out

        device = "cuda:0"
        input_shape = [2, 4096]
        # run torch to get the expected result
        data_torch = torch.randn(size=input_shape, device=device, dtype=torch.bfloat16) + 10
        data_torch.requires_grad = True
        torch_model = Model()
        torch_res = torch_model(input=data_torch)
        init_grad = torch.ones_like(torch_res)
        torch_res.backward(gradient=init_grad)
        # run ort
        ort_model = ORTModule(torch_model)
        data_ort = data_torch.detach().clone()
        data_ort.requires_grad = True
        ort_res = ort_model(input=data_ort)
        ort_res.backward(gradient=init_grad)
        # compare result
        torch.testing.assert_close(data_torch.grad, data_ort.grad, rtol=1e-5, atol=1e-4)


if __name__ == "__main__":
    unittest.main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import os
from itertools import product

import numpy as np
import scipy.special


def dtype_to_bytes(dtype):
    type_map = {
        "float8_e4m3fn": 1,
        "float8_e4m3fnuz": 1,
        "float8_e5m2": 1,
        "float8_e5m2fnuz": 1,
        "float16": 2,
        "float32": 4,
        "float64": 8,
    }
    return type_map[dtype]


def transab_to_suffix(transab):
    return {
        (True, True): "TT",
        (True, False): "TN",
        (False, True): "NT",
        (False, False): "NN",
    }[tuple(transab)]


def dtype_to_suffix(dtype):
    return {
        "float32": "float",
        "float16": "half",
        "float8_e4m3fn": "fp8e4m3fn",
        "float8_e4m3fnuz": "fp8e4m3fnuz",
    }[dtype]


def get_gemm_bound(
    dtype: str,
    a: np.ndarray,
    b: np.ndarray,
    c: np.ndarray,
    transa: bool,
    transb: bool,
    a_b_positive=False,  # if both a and b are positive matrix, we can skip coeff computation
):
    k = b.shape[1] if transb else b.shape[0]
    # The machine epsilon, unit roundoff, the smallest positive floating point number n such that the floating point
    # number that represents 1 + n is greater than 1.
    machine_eps = 2.0 ** -(24 if dtype == "float32" else 11)

    # The following implements error bound 5.7 in paper I. C. Ipsen and H. Zhou, “Probabilistic error analysis for
    # Inner Products,” SIAM Journal on Matrix Analysis and Applications, vol. 41, no. 4, pp. 1726-1741, 2020.
    # NOTE: the bound is not tight for float16 when k is large
    if a_b_positive:
        coeff = 1.0
    else:
        absa_mul_absb = np.abs(a.T if transa else a) @ np.abs(b.T if transb else b)
        coeff = np.max(absa_mul_absb / np.abs(c))
    gamma_2k = (1.0 + machine_eps) ** (2 * k) - 1.0
    bound_5_7 = coeff * np.sqrt(np.log(2 / 1e-10) * machine_eps * gamma_2k / 2)
    bound = bound_5_7

    return bound


def get_gemm_bert_sizes(full=True):
    bert_base_sizes = [
        # m, n, k
        (384, 768, 768),
        (384, 768, 768 * 3),
        (384, 768, 768 * 4),
        (384, 768 * 4, 768),
        (384, 1024, 1024),
        (384, 1024, 1024 * 3),
        (384, 1024, 1024 * 4),
        (384, 1024 * 4, 1024),
    ]

    # we then multiply m with the batch size
    if full:
        batch_sizes = [1, 64]
    else:
        batch_sizes = [1]
    bert_sizes = []
    for bsz in batch_sizes:
        bert_sizes.extend([(m * bsz, n, k) for m, n, k in bert_base_sizes])
    return bert_sizes


def get_gemm_basic_sizes(full=True):
    if full:
        return list(product([1, 3, 4, 16, 127, 128, 129, 133, 1024], repeat=3))

    # ck has various impls to be tested, use the full basic cases will result too many cases to test.
    # So we use a reduced combination here.
    return list(product([1, 4, 127, 133], [3, 16, 128], [3, 129, 1024]))


def softmax(x, *, is_log_softmax=False, axis=-1):
    x = x - np.max(x, axis=axis, keepdims=1)
    if is_log_softmax:
        return x - np.log(np.sum(np.exp(x), axis=axis, keepdims=1))
    return (np.exp(x)) / np.sum(np.exp(x), axis=axis, keepdims=1)


def _matmul(a, b):
    if os.getenv("KERNEL_EXPLORER_TEST_USE_CUPY", "0") == "1":
        import cupy as cp

        return (cp.asarray(a) @ cp.asarray(b)).get()
    else:
        return a @ b


def matmul(a, b, transa=False, transb=False):
    return _matmul(a.T if transa else a, b.T if transb else b)


def fast_gelu(x, bias):
    x = x + bias
    y = 0.5 * x * (1 + np.tanh(0.797885 * x + 0.035677 * x * x * x))
    return y


def gelu(x, bias):
    x = x + bias
    return 0.5 * x * (1 + scipy.special.erf(x / np.sqrt(2)))


def relu(x, bias):
    x = x + bias
    return np.max(x, 0, keepdims=True)


def root_mean_square(x, axis, epsilon):
    rms = np.sqrt(np.mean(np.square(x), axis=axis, keepdims=True) + epsilon)
    return rms


def standardization(x, axis, epsilon):
    mean = np.mean(x, axis=axis, keepdims=True)
    variance = np.var(x, axis=axis, keepdims=True)
    return (x - mean) / np.sqrt(variance + epsilon)



import onnx
from onnx import OperatorSetIdProto, TensorProto, helper

onnxdomain = OperatorSetIdProto()
onnxdomain.version = 12
# The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
onnxdomain.domain = ""
msdomain = OperatorSetIdProto()
msdomain.version = 1
msdomain.domain = "com.microsoft"
opsets = [onnxdomain, msdomain]


def save(model_path, nodes, inputs, outputs, initializers):
    graph = helper.make_graph(nodes, "TransposeGemmTest", inputs, outputs, initializers)

    model = helper.make_model(graph, opset_imports=opsets, producer_name="onnxruntime-test")

    onnx.save(model, model_path)


# (A')'B' = AB'
def gemm_transpose_2outputs_from_transpose(model_path):
    nodes = [
        helper.make_node("Transpose", ["A"], ["tp0"], "TransposeA"),
        helper.make_node("Transpose", ["B"], ["tp1"], "TransposeB"),
        helper.make_node("Gemm", ["tp0", "tp1"], ["output"], "Gemm", alpha=3.0, transA=1),
        helper.make_node("Identity", ["tp0"], ["output2"], "IdentityAt"),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, ["N", "K"]),
    ]

    outputs = [
        helper.make_tensor_value_info("output", TensorProto.FLOAT, ["M", "N"]),
        helper.make_tensor_value_info("output2", TensorProto.FLOAT, ["K", "M"]),
    ]

    save(model_path, nodes, inputs, outputs, [])


# (A')'B' = AB'  and  (B')'C = BC
def gemm_transpose_2outputs_from_transpose_to_2gemms(model_path):
    nodes = [
        helper.make_node("Transpose", ["A"], ["tp0"], "TransposeA"),
        helper.make_node("Transpose", ["B"], ["tp1"], "TransposeB"),
        helper.make_node("Gemm", ["tp0", "tp1"], ["output"], "Gemm1", alpha=3.0, transA=1),
        helper.make_node("Gemm", ["tp1", "C"], ["output3"], "Gemm2", alpha=3.0, transA=1),
        helper.make_node("Identity", ["tp0"], ["output2"], "IdentityAt"),
    ]

    inputs = [
        helper.make_tensor_value_info("A", TensorProto.FLOAT, ["M", "K"]),
        helper.make_tensor_value_info("B", TensorProto.FLOAT, ["N", "K"]),
        helper.make_tensor_value_info("C", TensorProto.FLOAT, ["K", "L"]),
    ]

    outputs = [
        helper.make_tensor_value_info("output", TensorProto.FLOAT, ["M", "N"]),
        helper.make_tensor_value_info("output2", TensorProto.FLOAT, ["K", "M"]),
        helper.make_tensor_value_info("output3", TensorProto.FLOAT, ["N", "L"]),
    ]

    save(model_path, nodes, inputs, outputs, [])


gemm_transpose_2outputs_from_transpose("gemm_transpose_2outputs_from_transpose.onnx")
gemm_transpose_2outputs_from_transpose_to_2gemms("gemm_transpose_2outputs_from_transpose_to_2gemms.onnx")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest

import numpy as np
from helper import get_name

import onnxruntime as onnxrt


class CudaGraphHelper:
    def __init__(
        self,
        ort_session: onnxrt.InferenceSession,
        input_and_output_shape: dict[str, list[int]],
        device_id: int = 0,
    ):
        self.input_names = [input.name for input in ort_session.get_inputs()]
        self.output_names = [output.name for output in ort_session.get_outputs()]

        self.input_and_output_shape = input_and_output_shape
        self.io_numpy_type = self.get_io_numpy_type_map(ort_session)
        self.io_binding = ort_session.io_binding()
        self.io_ort_value = {}

        for name in self.input_names + self.output_names:
            ort_value = onnxrt.OrtValue.ortvalue_from_shape_and_type(
                input_and_output_shape[name], self.io_numpy_type[name], "cuda", device_id
            )
            self.io_ort_value[name] = ort_value
            if name in self.input_names:
                self.io_binding.bind_ortvalue_input(name, ort_value)
            else:
                self.io_binding.bind_ortvalue_output(name, ort_value)

    def get_io_numpy_type_map(self, ort_session: onnxrt.InferenceSession):
        ort_type_to_numpy_type = {
            "tensor(int64)": np.longlong,
            "tensor(int32)": np.intc,
            "tensor(float)": np.float32,
            "tensor(float16)": np.float16,
        }

        name_to_numpy_type = {}
        for _input in ort_session.get_inputs():
            name_to_numpy_type[_input.name] = ort_type_to_numpy_type[_input.type]

        for output in ort_session.get_outputs():
            name_to_numpy_type[output.name] = ort_type_to_numpy_type[output.type]

        return name_to_numpy_type

    def update_inputs(self, inputs: dict[str, np.ndarray]):
        for input_name in self.input_names:
            self.io_ort_value[input_name].update_inplace(inputs[input_name])

    def get_output(self, output_name: str):
        return self.io_ort_value[output_name].numpy()


class TestInferenceSessionWithCudaGraph(unittest.TestCase):
    def test_ort_value_update_in_place(self):
        x0 = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=np.float32)
        ortvalue_cpu = onnxrt.OrtValue.ortvalue_from_numpy(x0)
        np.testing.assert_allclose(x0, ortvalue_cpu.numpy())

        x1 = np.array([[10.0, 20.0], [30.0, 40.0], [50.0, 60.0]], dtype=np.float32)
        ortvalue_cpu.update_inplace(x1)
        np.testing.assert_allclose(x1, ortvalue_cpu.numpy())

        if "CUDAExecutionProvider" in onnxrt.get_available_providers():
            ortvalue_gpu = onnxrt.OrtValue.ortvalue_from_numpy(x0, "cuda", 0)
            np.testing.assert_allclose(x0, ortvalue_gpu.numpy())

            ortvalue_gpu.update_inplace(x1)
            np.testing.assert_allclose(x1, ortvalue_gpu.numpy())

    def test_select_ep_to_run_cuda_graph(self):
        if "TensorrtExecutionProvider" in onnxrt.get_available_providers():
            providers = [("TensorrtExecutionProvider", {"trt_cuda_graph_enable": True})]
            self.run_model_with_cuda_graph(providers)
        elif "CUDAExecutionProvider" in onnxrt.get_available_providers():
            providers = [("CUDAExecutionProvider", {"enable_cuda_graph": True})]
            self.run_model_with_cuda_graph(providers)
            self.run_model_with_cuda_graph_annotation(providers)

    def run_model_with_cuda_graph(self, providers):
        INPUT_SIZE = 1280  # noqa: N806
        x = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]] * INPUT_SIZE, dtype=np.float32)
        y = np.array([[0.0], [0.0], [0.0]] * INPUT_SIZE, dtype=np.float32)
        x_ortvalue = onnxrt.OrtValue.ortvalue_from_numpy(x, "cuda", 0)
        y_ortvalue = onnxrt.OrtValue.ortvalue_from_numpy(y, "cuda", 0)

        onnxrt.set_default_logger_severity(0)
        session = onnxrt.InferenceSession(get_name("matmul_2.onnx"), providers=providers)
        io_binding = session.io_binding()

        # Bind the input and output
        io_binding.bind_ortvalue_input("X", x_ortvalue)
        io_binding.bind_ortvalue_output("Y", y_ortvalue)

        ro = onnxrt.RunOptions()

        # One regular run for the necessary memory allocation and cuda graph capturing
        session.run_with_iobinding(io_binding, ro)
        expected_y = np.array([[5.0], [11.0], [17.0]] * INPUT_SIZE, dtype=np.float32)
        np.testing.assert_allclose(expected_y, y_ortvalue.numpy(), rtol=1e-05, atol=1e-05)

        # After capturing, CUDA graph replay happens from this Run onwards
        session.run_with_iobinding(io_binding, ro)
        np.testing.assert_allclose(expected_y, y_ortvalue.numpy(), rtol=1e-05, atol=1e-05)

        # Update input and then replay CUDA graph
        x_ortvalue.update_inplace(
            np.array(
                [[10.0, 20.0], [30.0, 40.0], [50.0, 60.0]] * INPUT_SIZE,
                dtype=np.float32,
            )
        )
        session.run_with_iobinding(io_binding, ro)
        np.testing.assert_allclose(
            np.array([[50.0], [110.0], [170.0]] * INPUT_SIZE, dtype=np.float32),
            y_ortvalue.numpy(),
            rtol=1e-05,
            atol=1e-05,
        )

    def run_model_with_cuda_graph_annotation(self, providers):
        INPUT_SIZE = 1280  # noqa: N806

        x_base = [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]]
        y_base = [[0.0], [0.0], [0.0], [0.0]]
        expected_y_base = [[5.0], [11.0], [17.0], [23.0]]

        x_base_mul_10 = [[10.0, 20.0], [30.0, 40.0], [50.0, 60.0], [70.0, 80.0]]
        expected_y_base_mul_10 = [[50.0], [110.0], [170.0], [230.0]]

        test_num = 4

        x_ortvalues = []
        y_ortvalues = []
        for i in range(test_num):
            x = np.array(x_base[: i + 1][:] * INPUT_SIZE, dtype=np.float32)
            y = np.array(y_base[: i + 1][:] * INPUT_SIZE, dtype=np.float32)
            x_ortvalues.append(onnxrt.OrtValue.ortvalue_from_numpy(x, "cuda", 0))
            y_ortvalues.append(onnxrt.OrtValue.ortvalue_from_numpy(y, "cuda", 0))

        onnxrt.set_default_logger_severity(0)
        session = onnxrt.InferenceSession(get_name("matmul_2.onnx"), providers=providers)
        io_bindings = [session.io_binding()] * test_num
        ro = onnxrt.RunOptions()

        # Regular run to capture CUDA graph
        for i in range(test_num):
            io_bindings[i].bind_ortvalue_input("X", x_ortvalues[i])
            io_bindings[i].bind_ortvalue_output("Y", y_ortvalues[i])
            # TODO: Temporarily remove the default cuda graph capture test for the first regular run
            # because it fails on a training CI. Need to investigate the root cause.
            ro.add_run_config_entry("gpu_graph_id", str(i + 1))
            io_bindings[i].synchronize_inputs()
            session.run_with_iobinding(io_bindings[i], ro)
            io_bindings[i].synchronize_outputs()
            expected_y = np.array(expected_y_base[: i + 1][:] * INPUT_SIZE, dtype=np.float32)
            np.testing.assert_allclose(expected_y, y_ortvalues[i].numpy(), rtol=1e-05, atol=1e-05)

        del ro
        ro = onnxrt.RunOptions()

        # After capturing, CUDA graph replay happens from this Run onwards
        for i in range(test_num):
            # Update input and then replay CUDA graph
            x_ortvalues[i].update_inplace(np.array(x_base_mul_10[: i + 1][:] * INPUT_SIZE, dtype=np.float32))
            ro.add_run_config_entry("gpu_graph_id", str(i + 1))
            io_bindings[i].synchronize_inputs()
            session.run_with_iobinding(io_bindings[i], ro)
            io_bindings[i].synchronize_outputs()
            expected_y = np.array(expected_y_base_mul_10[: i + 1][:] * INPUT_SIZE, dtype=np.float32)
            np.testing.assert_allclose(expected_y, y_ortvalues[i].numpy(), rtol=1e-05, atol=1e-05)

    def test_arena_with_cuda_graph(self):
        if "CUDAExecutionProvider" in onnxrt.get_available_providers():
            # To test cuda graph catpure, we set Arena extend strategy to be SameAsRequested so as to detect any
            # potential memory allocation after the first run.
            providers = [
                ("CUDAExecutionProvider", {"enable_cuda_graph": True, "arena_extend_strategy": "kSameAsRequested"})
            ]
            test_model_path = get_name("squeezenet/model.onnx")

            input_and_output_shape = {
                "data_0": [16, 3, 224, 224],
                "softmaxout_1": [16, 1000, 1, 1],
            }

            session_options = onnxrt.SessionOptions()
            # It is optional to disable memory pattern since min_num_runs_before_cuda_graph_capture_ = 2.
            session_options.enable_mem_pattern = False
            session = onnxrt.InferenceSession(test_model_path, session_options, providers=providers)

            cuda_graph_helper = CudaGraphHelper(session, input_and_output_shape)
            io_binding = cuda_graph_helper.io_binding

            # Create a random input for testing.
            np.random.seed(0)
            inputs = {"data_0": np.random.randint(0, 256, size=input_and_output_shape["data_0"]).astype(np.float32)}

            # One regular run for the necessary memory allocation and cuda graph capturing
            cuda_graph_helper.update_inputs(inputs)
            session.run_with_iobinding(io_binding)
            expected_output = cuda_graph_helper.get_output("softmaxout_1")

            # After capturing, CUDA graph replay happens from this Run onwards
            cuda_graph_helper.update_inputs(inputs)
            session.run_with_iobinding(io_binding)
            output = cuda_graph_helper.get_output("softmaxout_1")

            np.testing.assert_allclose(expected_output, output, rtol=1e-02, atol=1e-02)


if __name__ == "__main__":
    unittest.main()



import argparse
import shutil
import sys
from pathlib import Path

import numpy as np

# set to the directory the ONNX Runtime repo is in
# `git checkout https://github.com/microsoft/onnxruntime.git` if needed.
ORT_ROOT_DIR = Path(__file__).parents[3]
SOLUTION_DIR = Path(__file__).parent

# add path for test data/dir generation utils
sys.path.append(str(ORT_ROOT_DIR / "tools" / "python"))


def parse_args():
    parser = argparse.ArgumentParser(
        description="""Setup the model and test data for usage with the MAUI model tester app.
        Input data will be randomly generated as needed.
        The model will be run locally and the output saved as expected output.
        Explicit input data or expected output data can be specified by providing .pb files with the input/output name
        and tensor. These can be created with /tools/python/onnx_test_data_utils.py.

        See https://github.com/microsoft/onnxruntime/blob/main/tools/python/PythonTools.md#creating-a-test-directory-for-a-model  # noqa
        for info on creating specific input or expected output""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--symbolic_dims",
        "-s",
        help="Symbolic dimension values if the model inputs have symbolic dimensions and the input data is being "
        "generated. Format is `name=value {name2=value2 ...}.",
        type=str,
        nargs="+",
        required=False,
    )

    parser.add_argument(
        "--input_data",
        "-i",
        help="Input data pb files created with onnx_test_data_utils.py. Multiple can be specified.",
        type=Path,
        nargs="+",
        required=False,
    )

    parser.add_argument(
        "--output_data",
        "-o",
        help="Expected output data pb files created with onnx_test_data_utils.py. Multiple can be specified.",
        type=Path,
        nargs="+",
        required=False,
    )

    parser.add_argument(
        "--model_path",
        "-m",
        help="Path to ONNX model to use. Model will be copied into the test app",
        type=Path,
        required=True,
    )

    args = parser.parse_args()

    args.model_path.resolve(strict=True)

    # convert symbolic dims to dictionary
    symbolic_dims = None
    if args.symbolic_dims:
        symbolic_dims = {}
        for value in args.symbolic_dims:
            pieces = value.split("=")
            assert len(pieces) == 2
            name = pieces[0].strip()
            dim_value = int(pieces[1].strip())
            symbolic_dims[name] = dim_value

    args.symbolic_dims = symbolic_dims

    return args


def create_existing_data_map(pb_files: list[Path]):
    import onnx_test_data_utils as data_utils

    data_map = {}
    for file in pb_files:
        file.resolve(strict=True)
        name, data = data_utils.read_tensorproto_pb_file(str(file))
        data_map[name] = data

    return data_map


def add_model_and_test_data_to_app(
    model_path: Path,
    symbolic_dims: dict[str, int] | None = None,
    input_map: dict[str, np.ndarray] | None = None,
    output_map: dict[str, np.ndarray] | None = None,
):
    import ort_test_dir_utils as utils

    output_path = SOLUTION_DIR / "Resources" / "Raw"
    test_name = "test_data"

    test_path = output_path / test_name
    # remove existing data
    if test_path.exists():
        shutil.rmtree(test_path)

    # If you want to directly create input data without using onnx_test_data_utils you can edit the input map here
    # if not input_map:
    #     input_map = {}
    #
    # input_map['Input3'] = np.random.rand(1, 1, 28, 28).astype(np.float32)

    utils.create_test_dir(
        str(model_path),
        str(output_path),
        test_name,
        # Explicit input data. Any missing required inputs will have data generated for them.
        name_input_map=input_map,
        # Optional map for any symbolic values.
        symbolic_dim_values_map=symbolic_dims,
        # Expected output can be provided if you want to validate model output against this.
        name_output_map=output_map,
    )

    # create_test_dir will copy the model to the output directory.
    # rename the copied model to the generic name the app expects.
    copied_model = output_path / test_name / model_path.name
    copied_model.rename(copied_model.with_name("model.onnx"))

    # add a text file with the original model path just so there's some info on where it came from
    with open(test_path / "model_info.txt", "w") as model_info_file:
        model_info_file.write(str(model_path))


def create_test_data():
    args = parse_args()

    input_map = None
    output_map = None

    if args.input_data:
        input_map = create_existing_data_map(args.input_data)

    if args.output_data:
        output_map = create_existing_data_map(args.output_data)

    add_model_and_test_data_to_app(args.model_path, args.symbolic_dims, input_map, output_map)


if __name__ == "__main__":
    create_test_data()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# This tool generates a tiny GPT2 model for testing fusion script.
# You can use benchmark_gpt2.py to get a gpt2 ONNX model as input of this tool.

import argparse
import os
from pathlib import Path

import numpy as np
import onnx
import onnx.utils
from onnx import ModelProto, TensorProto, numpy_helper
from onnxruntime_tools.transformers.onnx_model import OnnxModel

import onnxruntime

DICT_SIZE = 20
SEQ_LEN = 5
""" This class creates a tiny bert model for test purpose. """

# parameters of input base model.
old_parameters = {
    "seq_len": 5,
    "hidden_size": 768,
    "num_heads": 12,
    "size_per_head": 64,
    "word_dict_size": [50257],  # list of supported dictionary size.
    "max_word_position": 1024,
}

# parameters of output tiny model.
new_parameters = {
    "seq_len": SEQ_LEN,
    "hidden_size": 4,
    "num_heads": 2,
    "size_per_head": 2,
    "word_dict_size": DICT_SIZE,
    "max_word_position": 8,
}


class TinyGpt2Model(OnnxModel):
    def __init__(self, model):
        super().__init__(model)
        self.resize_model()

    def resize_weight(self, initializer_name, target_shape):
        weight = self.get_initializer(initializer_name)
        w = numpy_helper.to_array(weight)

        target_w = w
        if len(target_shape) == 1:
            target_w = w[: target_shape[0]]
        elif len(target_shape) == 2:
            target_w = w[: target_shape[0], : target_shape[1]]
        elif len(target_shape) == 3:
            target_w = w[: target_shape[0], : target_shape[1], : target_shape[2]]
        elif len(target_shape) == 4:
            target_w = w[
                : target_shape[0],
                : target_shape[1],
                : target_shape[2],
                : target_shape[3],
            ]
        else:
            print("at most 3 dimensions")

        tensor = onnx.helper.make_tensor(
            name=initializer_name + "_resize",
            data_type=TensorProto.FLOAT,
            dims=target_shape,
            vals=target_w.flatten().tolist(),
        )

        return tensor

    def resize_model(self):
        graph = self.model.graph
        initializers = graph.initializer

        for input in graph.input:
            if input.type.tensor_type.shape.dim[1].dim_value == old_parameters["seq_len"]:
                print("input", input.name, input.type.tensor_type.shape)
                input.type.tensor_type.shape.dim[1].dim_value = new_parameters["seq_len"]
                print("=>", input.type.tensor_type.shape)

        reshapes = {}
        for initializer in initializers:
            tensor = numpy_helper.to_array(initializer)
            if initializer.data_type == TensorProto.FLOAT:
                dtype = np.float32
            elif initializer.data_type == TensorProto.INT32:
                dtype = np.int32
            elif initializer.data_type == TensorProto.INT64:
                dtype = np.int64
            else:
                print("data type not supported by this tool:", dtype)

            if len(tensor.shape) == 1 and tensor.shape[0] == 1:
                if tensor == old_parameters["num_heads"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["num_heads"],
                        "=>[",
                        new_parameters["num_heads"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([new_parameters["num_heads"]], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["seq_len"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["seq_len"],
                        "=>[",
                        new_parameters["seq_len"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([new_parameters["seq_len"]], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["size_per_head"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["size_per_head"],
                        "=>[",
                        new_parameters["size_per_head"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([new_parameters["size_per_head"]], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["hidden_size"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["hidden_size"],
                        "=>[",
                        new_parameters["hidden_size"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([new_parameters["hidden_size"]], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == 4 * old_parameters["hidden_size"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        4 * old_parameters["hidden_size"],
                        "=>[",
                        4 * new_parameters["hidden_size"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([4 * new_parameters["hidden_size"]], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == 3 * old_parameters["hidden_size"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        3 * old_parameters["hidden_size"],
                        "=>[",
                        3 * new_parameters["hidden_size"],
                        "]",
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray([3 * new_parameters["hidden_size"]], dtype=dtype),
                            initializer.name,
                        )
                    )
            elif len(tensor.shape) == 0:
                if tensor == old_parameters["num_heads"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["num_heads"],
                        "=>",
                        new_parameters["num_heads"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(new_parameters["num_heads"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["seq_len"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["seq_len"],
                        "=>",
                        new_parameters["seq_len"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(new_parameters["seq_len"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["size_per_head"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["size_per_head"],
                        "=>",
                        new_parameters["size_per_head"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(new_parameters["size_per_head"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == old_parameters["hidden_size"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        old_parameters["hidden_size"],
                        "=>",
                        new_parameters["hidden_size"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(new_parameters["hidden_size"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == 4 * old_parameters["hidden_size"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        4 * old_parameters["hidden_size"],
                        "=>",
                        4 * new_parameters["hidden_size"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(4 * new_parameters["hidden_size"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == 3 * old_parameters["hidden_size"]:
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        3 * old_parameters["hidden_size"],
                        "=>",
                        3 * new_parameters["hidden_size"],
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(3 * new_parameters["hidden_size"], dtype=dtype),
                            initializer.name,
                        )
                    )
                elif tensor == 1.0 / np.sqrt(old_parameters["size_per_head"]):
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        1.0 / np.sqrt(old_parameters["size_per_head"]),
                        "=>",
                        1.0 / np.sqrt(new_parameters["size_per_head"]),
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(
                                1.0 / np.sqrt(new_parameters["size_per_head"]),
                                dtype=dtype,
                            ),
                            initializer.name,
                        )
                    )
                elif tensor == np.sqrt(old_parameters["size_per_head"]):
                    print(
                        f"initializer type={initializer.data_type}",
                        initializer.name,
                        np.sqrt(old_parameters["size_per_head"]),
                        "=>",
                        np.sqrt(new_parameters["size_per_head"]),
                    )
                    initializer.CopyFrom(
                        numpy_helper.from_array(
                            np.asarray(np.sqrt(new_parameters["size_per_head"]), dtype=dtype),
                            initializer.name,
                        )
                    )

            new_shape = []
            shape_changed = False
            for dim in tensor.shape:
                if dim == old_parameters["hidden_size"]:
                    new_shape.append(new_parameters["hidden_size"])
                    shape_changed = True
                elif dim == 4 * old_parameters["hidden_size"]:
                    new_shape.append(4 * new_parameters["hidden_size"])
                    shape_changed = True
                elif dim == 3 * old_parameters["hidden_size"]:
                    new_shape.append(3 * new_parameters["hidden_size"])
                    shape_changed = True
                elif dim in old_parameters["word_dict_size"]:
                    new_shape.append(new_parameters["word_dict_size"])
                    shape_changed = True
                elif dim == old_parameters["max_word_position"]:
                    new_shape.append(new_parameters["max_word_position"])
                    shape_changed = True
                else:
                    new_shape.append(dim)
            if shape_changed:
                reshapes[initializer.name] = new_shape
                print("initializer", initializer.name, tensor.shape, "=>", new_shape)

        for initializer_name, reshape_name in reshapes.items():
            self.replace_input_of_all_nodes(initializer_name, initializer_name + "_resize")
            tensor = self.resize_weight(initializer_name, reshape_name)
            self.model.graph.initializer.extend([tensor])

        # Add node name, replace split node attribute.
        nodes_to_add = []
        nodes_to_remove = []
        for i, node in enumerate(graph.node):
            if node.op_type == "Split":
                nodes_to_add.append(
                    onnx.helper.make_node(
                        "Split",
                        node.input,
                        node.output,
                        name=f"Split_{i}",
                        axis=2,
                        split=[
                            new_parameters["hidden_size"],
                            new_parameters["hidden_size"],
                            new_parameters["hidden_size"],
                        ],
                    )
                )
                nodes_to_remove.append(node)
                print(
                    "update split",
                    [
                        new_parameters["hidden_size"],
                        new_parameters["hidden_size"],
                        new_parameters["hidden_size"],
                    ],
                )
            if node.op_type == "Constant":
                for att in node.attribute:
                    if att.name == "value":
                        if numpy_helper.to_array(att.t) == old_parameters["num_heads"]:
                            nodes_to_add.append(
                                onnx.helper.make_node(
                                    "Constant",
                                    inputs=node.input,
                                    outputs=node.output,
                                    value=onnx.helper.make_tensor(
                                        name=att.t.name,
                                        data_type=TensorProto.INT64,
                                        dims=[],
                                        vals=[new_parameters["num_heads"]],
                                    ),
                                )
                            )
                            print(
                                "constant",
                                att.t.name,
                                old_parameters["num_heads"],
                                "=>",
                                new_parameters["num_heads"],
                            )
                        if numpy_helper.to_array(att.t) == np.sqrt(old_parameters["size_per_head"]):
                            nodes_to_add.append(
                                onnx.helper.make_node(
                                    "Constant",
                                    inputs=node.input,
                                    outputs=node.output,
                                    value=onnx.helper.make_tensor(
                                        name=att.t.name,
                                        data_type=TensorProto.FLOAT,
                                        dims=[],
                                        vals=[np.sqrt(new_parameters["size_per_head"])],
                                    ),
                                )
                            )
                            print(
                                "constant",
                                att.t.name,
                                np.sqrt(old_parameters["size_per_head"]),
                                "=>",
                                np.sqrt(new_parameters["size_per_head"]),
                            )
            else:
                node.name = node.op_type + "_" + str(i)
        for node in nodes_to_remove:
            graph.node.remove(node)
        graph.node.extend(nodes_to_add)

        for i, input in enumerate(self.model.graph.input):
            if i > 0:
                dim_proto = input.type.tensor_type.shape.dim[2]
                dim_proto.dim_value = new_parameters["num_heads"]
                dim_proto = input.type.tensor_type.shape.dim[4]
                dim_proto.dim_value = new_parameters["size_per_head"]

        for i, output in enumerate(self.model.graph.output):
            if i == 0:
                dim_proto = output.type.tensor_type.shape.dim[2]
                dim_proto.dim_value = new_parameters["hidden_size"]
            if i > 0:
                dim_proto = output.type.tensor_type.shape.dim[2]
                dim_proto.dim_value = new_parameters["num_heads"]
                dim_proto = output.type.tensor_type.shape.dim[4]
                dim_proto.dim_value = new_parameters["size_per_head"]


def generate_test_data(
    onnx_file,
    output_path,
    batch_size=1,
    use_cpu=True,
    input_tensor_only=False,
    dictionary_size=DICT_SIZE,
    test_cases=1,
    output_optimized_model=False,
):
    for test_case in range(test_cases):
        sequence_length = 3
        input_1 = np.random.randint(dictionary_size, size=(batch_size, 1), dtype=np.int64)
        tensor_1 = numpy_helper.from_array(input_1, "input_ids")

        path = os.path.join(output_path, "test_data_set_" + str(test_case))
        try:
            os.mkdir(path)
        except OSError:
            print(f"Creation of the directory {path} failed")
        else:
            print(f"Successfully created the directory {path} ")

        sess_options = onnxruntime.SessionOptions()
        sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_DISABLE_ALL
        sess = onnxruntime.InferenceSession(onnx_file, sess_options, providers=["CPUExecutionProvider"])

        input1_name = sess.get_inputs()[0].name
        output_names = [output.name for output in sess.get_outputs()]
        inputs = {input1_name: input_1}

        with open(os.path.join(path, f"input_{0}.pb"), "wb") as f:
            f.write(tensor_1.SerializeToString())

        for i in range(12):
            input_name = f"past_{i}"
            input = np.random.rand(
                2,
                batch_size,
                new_parameters["num_heads"],
                sequence_length,
                new_parameters["size_per_head"],
            ).astype(np.float32)
            tensor = numpy_helper.from_array(input, input_name)
            inputs.update({input_name: input})

            with open(os.path.join(path, f"input_{1 + i}.pb"), "wb") as f:
                f.write(tensor.SerializeToString())

        if input_tensor_only:
            return

        result = sess.run(output_names, inputs)
        print("result 0 shape:", result[0].shape)
        print("result 1 shape:", result[1].shape)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", required=True, type=str)
    parser.add_argument("--output", required=True, type=str)
    parser.add_argument("--output_optimized_model", required=False, action="store_true")
    parser.set_defaults(output_optimized_model=False)
    args = parser.parse_args()

    model = ModelProto()
    with open(args.input, "rb") as f:
        model.ParseFromString(f.read())

    bert_model = TinyGpt2Model(model)

    bert_model.update_graph()
    bert_model.remove_unused_constant()

    print("opset version", bert_model.model.opset_import[0].version)

    with open(args.output, "wb") as out:
        out.write(bert_model.model.SerializeToString())

    p = Path(args.output)
    data_path = p.parent

    generate_test_data(
        args.output,
        data_path,
        batch_size=1,
        use_cpu=True,
        output_optimized_model=args.output_optimized_model,
    )


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import math
import os
from types import ModuleType

import torch

from .._cache import ModuleCache, PyCodeCache
from .._utils import gen_unique_name, next_power_of_2

_DEBUG_MODE = "ORTMODULE_TRITON_DEBUG" in os.environ and int(os.getenv("ORTMODULE_TRITON_DEBUG")) == 1


_MM_TEMPLATE = """
import torch
import triton
import triton.language as tl


@triton.autotune(
    configs=[
{autotune_configs}
    ],
    key=["M", "N", "K"],
)
@triton.jit
def {kernel_name}(
    A, B, OUT, M, N, K, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr, BLOCK_K: tl.constexpr, GROUP_M: tl.constexpr
):
    M = {M}
    N = {N}
    K = {K}
    stride_am = {stride_am}
    stride_ak = {stride_ak}
    stride_bk = {stride_bk}
    stride_bn = {stride_bn}

    # based on triton.ops.matmul
    pid = tl.program_id(0)
    grid_m = (M + BLOCK_M - 1) // BLOCK_M
    grid_n = (N + BLOCK_N - 1) // BLOCK_N

    # re-order program ID for better L2 performance
    width = GROUP_M * grid_n
    group_id = pid // width
    group_size = min(grid_m - group_id * GROUP_M, GROUP_M)
    pid_m = group_id * GROUP_M + (pid % group_size)
    pid_n = (pid % width) // (group_size)

    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    ram = tl.max_contiguous(tl.multiple_of(rm % M, BLOCK_M), BLOCK_M)
    rbn = tl.max_contiguous(tl.multiple_of(rn % N, BLOCK_N), BLOCK_N)
    rk = tl.arange(0, BLOCK_K)
    A = A + (ram[:, None] * stride_am + rk[None, :] * stride_ak)
    B = B + (rk[:, None] * stride_bk + rbn[None, :] * stride_bn)

    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)
    for k in range(K, 0, -BLOCK_K):
        if {even_k}:
            a = tl.load(A)
            b = tl.load(B)
        else:
            a = tl.load(A, mask=rk[None, :] < k, other=0.)
            b = tl.load(B, mask=rk[:, None] < k, other=0.)
        acc += tl.dot(a, b, allow_tf32={allow_tf32})
        A += BLOCK_K * stride_ak
        B += BLOCK_K * stride_bk

    # rematerialize rm and rn to save registers
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    idx_m = rm[:, None]
    idx_n = rn[None, :]
    mask = (idx_m < M) & (idx_n < N)
    OUT = OUT + (idx_m * N + idx_n)
    tl.store(OUT, {post_process}, mask=mask)


def {func_name}(a, b, out):
    # 1D launch kernel where each block gets its own program.
    grid = lambda META: (triton.cdiv({M}, META["BLOCK_M"]) * triton.cdiv({N}, META["BLOCK_N"]),)
    {kernel_name}[grid](a, b, out, {M}, {N}, {K})
"""


_BMM_TEMPLATE = """
import torch
import triton
import triton.language as tl


@triton.autotune(
    configs=[
{autotune_configs}
    ],
    key=["M", "N", "K"],
)
@triton.jit
def {kernel_name}(
    A, B, OUT, M, N, K, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr, BLOCK_K: tl.constexpr, GROUP_M: tl.constexpr
):
    M = {M}
    N = {N}
    K = {K}
    stride_aq = {stride_aq}
    stride_am = {stride_am}
    stride_ak = {stride_ak}
    stride_bq = {stride_bq}
    stride_bk = {stride_bk}
    stride_bn = {stride_bn}
    stride_cq = M * N

    # based on triton.ops.matmul
    pid = tl.program_id(0)
    grid_m = (M + BLOCK_M - 1) // BLOCK_M
    grid_n = (N + BLOCK_N - 1) // BLOCK_N

    # re-order program ID for better L2 performance
    width = GROUP_M * grid_n
    group_id = pid // width
    group_size = min(grid_m - group_id * GROUP_M, GROUP_M)
    pid_m = group_id * GROUP_M + (pid % group_size)
    pid_n = (pid % width) // (group_size)

    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    ram = tl.max_contiguous(tl.multiple_of(rm % M, BLOCK_M), BLOCK_M)
    rbn = tl.max_contiguous(tl.multiple_of(rn % N, BLOCK_N), BLOCK_N)
    rk = tl.arange(0, BLOCK_K)

    idx_q = tl.program_id(1)  # batch dimension for BMM
    A = A + (ram[:, None] * stride_am + rk[None, :] * stride_ak + idx_q * stride_aq)
    B = B + (rk[:, None] * stride_bk + rbn[None, :] * stride_bn + idx_q * stride_bq)

    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)
    for k in range(K, 0, -BLOCK_K):
        if {even_k}:
            a = tl.load(A)
            b = tl.load(B)
        else:
            a = tl.load(A, mask=rk[None, :] < k, other=0.)
            b = tl.load(B, mask=rk[:, None] < k, other=0.)
        acc += tl.dot(a, b, allow_tf32={allow_tf32})
        A += BLOCK_K * stride_ak
        B += BLOCK_K * stride_bk

    # rematerialize rm and rn to save registers
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    idx_q = tl.program_id(1)  # batch dimension for BMM
    idx_m = rm[:, None]
    idx_n = rn[None, :]
    mask = (idx_m < M) & (idx_n < N)
    OUT = OUT + (idx_m * N + idx_n + idx_q * stride_cq)
    tl.store(OUT, {post_process}, mask=mask)


def {func_name}(a, b, out):
    # 1D launch kernel where each block gets its own program.
    grid = lambda META: (triton.cdiv({M}, META["BLOCK_M"]) * triton.cdiv({N}, META["BLOCK_N"]), {batch}, 1)
    {kernel_name}[grid](a, b, out, {M}, {N}, {K})
"""


_GEMM_TEMPLATE = """
import torch
import triton
import triton.language as tl


@triton.autotune(
    configs=[
{autotune_configs}
    ],
    key=["M", "N", "K"],
)
@triton.jit
def {kernel_name}(
    A, B, C, OUT, M, N, K, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr, BLOCK_K: tl.constexpr, GROUP_M: tl.constexpr
):
    M = {M}
    N = {N}
    K = {K}
    stride_am = {stride_am}
    stride_ak = {stride_ak}
    stride_bk = {stride_bk}
    stride_bn = {stride_bn}
    stride_cm = {stride_cm}
    stride_cn = {stride_cn}

    # based on triton.ops.matmul
    pid = tl.program_id(0)
    grid_m = (M + BLOCK_M - 1) // BLOCK_M
    grid_n = (N + BLOCK_N - 1) // BLOCK_N

    # re-order program ID for better L2 performance
    width = GROUP_M * grid_n
    group_id = pid // width
    group_size = min(grid_m - group_id * GROUP_M, GROUP_M)
    pid_m = group_id * GROUP_M + (pid % group_size)
    pid_n = (pid % width) // (group_size)

    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    ram = tl.max_contiguous(tl.multiple_of(rm % M, BLOCK_M), BLOCK_M)
    rbn = tl.max_contiguous(tl.multiple_of(rn % N, BLOCK_N), BLOCK_N)
    rk = tl.arange(0, BLOCK_K)
    A = A + (ram[:, None] * stride_am + rk[None, :] * stride_ak)
    B = B + (rk[:, None] * stride_bk + rbn[None, :] * stride_bn)

    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)
    for k in range(K, 0, -BLOCK_K):
        if {even_k}:
            a = tl.load(A)
            b = tl.load(B)
        else:
            a = tl.load(A, mask=rk[None, :] < k, other=0.)
            b = tl.load(B, mask=rk[:, None] < k, other=0.)
        acc += tl.dot(a, b, allow_tf32={allow_tf32})
        A += BLOCK_K * stride_ak
        B += BLOCK_K * stride_bk

    # rematerialize rm and rn to save registers
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    idx_m = rm[:, None]
    idx_n = rn[None, :]
    mask = (idx_m < M) & (idx_n < N)
    C = C + (idx_m * stride_cm + idx_n * stride_cn)
    c = tl.load(C, mask=mask, other=0.)
    OUT = OUT + (idx_m * N + idx_n)
    tl.store(OUT, {post_process} + c * {beta}, mask=mask)


def {func_name}(a, b, c, out):
    # 1D launch kernel where each block gets its own program.
    grid = lambda META: (triton.cdiv({M}, META["BLOCK_M"]) * triton.cdiv({N}, META["BLOCK_N"]),)
    {kernel_name}[grid](a, b, c, out, {M}, {N}, {K})
"""


def _mm_configs(dtype, m, n, k, trans_a, trans_b, alpha, func_name):
    condidates = [
        # "BLOCK_M", "BLOCK_N", "BLOCK_K", "num_stages", "num_warps"
        (32, 32, 16, 1, 2),
        (64, 64, 16, 2, 4),
        (64, 64, 32, 2, 4),
        (64, 128, 32, 3, 4),
        (128, 64, 32, 3, 4),
        (64, 128, 32, 4, 8),
        (128, 64, 32, 4, 8),
        (64, 32, 32, 5, 8),
        (32, 64, 32, 5, 8),
        (128, 128, 32, 2, 8),
        (64, 64, 64, 3, 8),
        (32, 32, 128, 2, 4),
    ]
    tm = max(next_power_of_2(m), 16)
    tn = max(next_power_of_2(n), 16)
    tk = max(next_power_of_2(k), 16)
    config_set = set()
    config_strs = []
    max_bk = 1
    for bm, bn, bk, num_stages, num_warps in condidates:
        new_bm = min(bm, tm)
        new_bn = min(bn, tn)
        new_bk = min(bk, tk)
        if (new_bm, new_bn, new_bk) in config_set:
            continue
        config_set.add((new_bm, new_bn, new_bk))
        if new_bk > max_bk:
            max_bk = new_bk
        new_num_warps = min(num_warps, new_bm * new_bn // 256)
        config_strs.append(
            f'        triton.Config({{"BLOCK_M": {new_bm}, "BLOCK_N": {new_bn}, "BLOCK_K": {new_bk}, "GROUP_M": 8}}, '
            f"num_stages={num_stages}, num_warps={new_num_warps}),"
        )
    autotune_configs = "\n".join(config_strs)
    post_process = "acc" if alpha == 1.0 else f"acc * {alpha}"
    if dtype == torch.float16:
        if alpha != 1.0:
            post_process = f"({post_process})"
        post_process = f"{post_process}.to(tl.float16)"
    return {
        "autotune_configs": autotune_configs,
        "kernel_name": f"kernel_{func_name}",
        "M": m,
        "N": n,
        "K": k,
        "stride_am": (1 if trans_a else k),
        "stride_ak": (m if trans_a else 1),
        "stride_bk": (1 if trans_b else n),
        "stride_bn": (k if trans_b else 1),
        "even_k": (k % max_bk == 0),
        "allow_tf32": torch.backends.cuda.matmul.allow_tf32,
        "post_process": post_process,
        "func_name": func_name,
    }


def _gen_mm_key(dtype: torch.dtype, m: int, n: int, k: int, trans_a: bool, trans_b: bool, alpha: float) -> int:
    return hash(f"mm|{dtype}|{m}|{n}|{k}|{trans_a}|{trans_b}|{alpha}")


def _gen_mm_module(
    dtype: torch.dtype, m: int, n: int, k: int, trans_a: bool, trans_b: bool, alpha: float
) -> tuple[str, ModuleType]:
    func_name = gen_unique_name("mm")
    kwargs = _mm_configs(dtype, m, n, k, trans_a, trans_b, alpha, func_name)
    src_code = _MM_TEMPLATE.format(**kwargs)
    if _DEBUG_MODE:
        os.makedirs(os.path.dirname("triton_debug/"), exist_ok=True)
        with open(f"triton_debug/{func_name}.py", "w", encoding="utf-8") as f:
            f.write(src_code)
    return func_name, PyCodeCache().load(src_code)


def _gen_gemm_key(
    dtype: torch.dtype,
    m: int,
    n: int,
    k: int,
    stride_cm: int,
    stride_cn: int,
    trans_a: bool,
    trans_b: bool,
    alpha: float,
    beta: float,
) -> int:
    return hash(f"gemm|{dtype}|{m}|{n}|{k}|{stride_cm}|{stride_cn}|{trans_a}|{trans_b}|{alpha}|{beta}")


def _gen_gemm_module(
    dtype: torch.dtype,
    m: int,
    n: int,
    k: int,
    stride_cm: int,
    stride_cn: int,
    trans_a: bool,
    trans_b: bool,
    alpha: float,
    beta: float,
) -> tuple[str, ModuleType]:
    func_name = gen_unique_name("gemm")
    kwargs = _mm_configs(dtype, m, n, k, trans_a, trans_b, alpha, func_name)
    kwargs["stride_cm"] = stride_cm
    kwargs["stride_cn"] = stride_cn
    kwargs["beta"] = beta
    src_code = _GEMM_TEMPLATE.format(**kwargs)
    if _DEBUG_MODE:
        os.makedirs(os.path.dirname("triton_debug/"), exist_ok=True)
        with open(f"triton_debug/{func_name}.py", "w", encoding="utf-8") as f:
            f.write(src_code)
    return func_name, PyCodeCache().load(src_code)


def _gen_bmm_key(
    dtype: torch.dtype, m: int, n: int, k: int, batch_a: int, batch_b: int, trans_a: bool, trans_b: bool, alpha: float
) -> int:
    return hash(f"bmm|{dtype}|{m}|{n}|{k}|{batch_a}|{batch_b}|{trans_a}|{trans_b}|{alpha}")


def _gen_bmm_module(
    dtype: torch.dtype, m: int, n: int, k: int, batch_a: int, batch_b: int, trans_a: bool, trans_b: bool, alpha: float
) -> tuple[str, ModuleType]:
    func_name = gen_unique_name("bmm")
    kwargs = _mm_configs(dtype, m, n, k, trans_a, trans_b, alpha, func_name)
    batch = max(batch_a, batch_b)
    kwargs["stride_aq"] = m * k if batch_a == batch else 0
    kwargs["stride_bq"] = k * n if batch_b == batch else 0
    kwargs["batch"] = batch
    src_code = _BMM_TEMPLATE.format(**kwargs)
    if _DEBUG_MODE:
        os.makedirs(os.path.dirname("triton_debug/"), exist_ok=True)
        with open(f"triton_debug/{func_name}.py", "w", encoding="utf-8") as f:
            f.write(src_code)
    return func_name, PyCodeCache().load(src_code)


def _matmul_internal(a, b, out, **kwargs):
    rank_a = len(a.shape)
    rank_b = len(b.shape)
    assert rank_a >= 2 and rank_b >= 2
    trans_a = kwargs.get("trans_a", False)
    trans_b = kwargs.get("trans_b", False)
    alpha = kwargs.get("alpha", 1.0)
    m = a.shape[-1] if trans_a else a.shape[-2]
    k = a.shape[-2] if trans_a else a.shape[-1]
    bk = b.shape[-1] if trans_b else b.shape[-2]
    assert k == bk
    n = b.shape[-2] if trans_b else b.shape[-1]
    assert rank_a == 2 or rank_b == 2 or a.shape[:-2] == b.shape[:-2]
    batch_shape = a.shape[:-2] if rank_a >= 3 else b.shape[:-2]
    if out is None:
        out = torch.empty((*batch_shape, m, n), dtype=a.dtype, device=a.device)
    else:
        assert out.shape == (*batch_shape, m, n)
    batch = math.prod(batch_shape)
    dtype = a.dtype
    if batch == 1 or (rank_a >= 3 and not trans_a and rank_b == 2):
        if batch != 1:
            m = batch * m
        func_name, mod = ModuleCache.load(_gen_mm_key, _gen_mm_module, dtype, m, n, k, trans_a, trans_b, alpha)
    else:
        batch_a = batch if rank_a >= 3 else 1
        batch_b = batch if rank_b >= 3 else 1
        func_name, mod = ModuleCache.load(
            _gen_bmm_key, _gen_bmm_module, dtype, m, n, k, batch_a, batch_b, trans_a, trans_b, alpha
        )
    func = getattr(mod, func_name)
    func(a, b, out)
    return out


def triton_matmul(a, b, **kwargs):
    """
    Compute matrix multiplication of two tensors.
    If trans_a is True in kwargs, input a will be transposed on last two dimensions before multiplication.
    If trans_b is True in kwargs, input b will be transposed on last two dimensions before multiplication.
    If alpha is specified in kwargs, the product will be multiplied by alpha.
    The input tensors can be of rank 2 or higher, but currently support only simple broadcasting on batch dimensions.
    I.e., the batch dimensions of a and b must either be equal, or one of them must be 1.
    """

    return _matmul_internal(a, b, None, **kwargs)


def triton_matmul_out(a, b, out, **kwargs):
    """
    Same as triton_matmul, except that the output is allocated and passed from outside.
    """

    _matmul_internal(a, b, out, **kwargs)


def _gemm_internal(a, b, c, out, **kwargs):
    assert len(a.shape) == 2 and len(b.shape) == 2
    trans_a = kwargs.get("trans_a", False)
    trans_b = kwargs.get("trans_b", False)
    alpha = kwargs.get("alpha", 1.0)
    beta = kwargs.get("beta", 1.0)
    m = a.shape[-1] if trans_a else a.shape[-2]
    k = a.shape[-2] if trans_a else a.shape[-1]
    bk = b.shape[-1] if trans_b else b.shape[-2]
    assert k == bk
    n = b.shape[-2] if trans_b else b.shape[-1]
    stride_cm = c.shape[-1] if len(c.shape) == 2 and c.shape[-2] == m else 0
    stride_cn = 1 if len(c.shape) >= 1 and c.shape[-1] == n else 0
    if out is None:
        out = torch.empty(m, n, dtype=a.dtype, device=a.device)
    else:
        assert out.shape == (m, n)
    dtype = a.dtype
    func_name, mod = ModuleCache.load(
        _gen_gemm_key, _gen_gemm_module, dtype, m, n, k, stride_cm, stride_cn, trans_a, trans_b, alpha, beta
    )
    func = getattr(mod, func_name)
    func(a, b, c, out)
    return out


def triton_gemm(a, b, c, **kwargs):
    """
    Compute alpha * a @ b + beta * c. Here a and b are 2D tensors, and c is broadcastable to the result.
    If trans_a is True in kwargs, input a will be transposed before multiplication.
    If trans_b is True in kwargs, input b will be transposed before multiplication.
    """

    return _gemm_internal(a, b, c, None, **kwargs)


def triton_gemm_out(a, b, c, out, **kwargs):
    """
    Same as triton_gemm, except that the output is allocated and passed from outside.
    """

    _gemm_internal(a, b, c, out, **kwargs)
